from __future__ import annotations
from collections.abc import Iterable
from collections import UserList
import sys
import warnings
from pathlib import Path
from copy import deepcopy
from typing import (
    Iterator,
    Union,
    Generator,
    ClassVar,
    Any,
    List,
    Dict,
    Sequence,
    TypeVar,
    get_type_hints,
    get_origin,
    get_args,
    Optional,
)
from functools import partial
from dataclasses import dataclass, fields, is_dataclass, MISSING
from collections import UserDict, UserList

from quam.serialisation import AbstractSerialiser, JSONSerialiser
from quam.utils import (
    get_dataclass_attr_annotations,
    ReferenceClass,
    string_reference,
    get_full_class_path,
    type_is_optional,
    generate_config_final_actions,
)
from quam.core.quam_instantiation import instantiate_quam_class
from quam.config import get_quam_config
from quam.utils.exceptions import InvalidReferenceError

from .qua_config_template import qua_config_template

from qm.type_hinting import DictQuaConfig

__all__ = [
    "QuamBase",
    "QuamRoot",
    "QuamComponent",
    "QuamDict",
    "QuamList",
    "quam_dataclass",
]


def _get_value_annotation(cls_or_obj: Union[type, object], attr: str) -> Optional[type]:
    """Get the type annotation for the values in a QuamDict or QuamList.

    If the QuamDict is defined as Dict[str, int], this will return int.
    If the QuamList is defined as List[int], this will return int.
    In all other cases, this will return None.
    """
    if cls_or_obj is None or attr is None:
        return None

    cls = cls_or_obj if isinstance(cls_or_obj, type) else cls_or_obj.__class__

    annotated_attrs = get_type_hints(cls)
    if attr not in annotated_attrs:
        return None

    attr_annotation = annotated_attrs[attr]
    origin = get_origin(attr_annotation)
    if origin == dict:
        return get_args(attr_annotation)[1]
    elif origin == list:
        return get_args(attr_annotation)[0]
    return None


def convert_dict_and_list(value, cls_or_obj=None, attr=None):
    """Convert a dict or list to a QuamDict or QuamList if possible."""
    if isinstance(value, dict):
        value_annotation = _get_value_annotation(cls_or_obj=cls_or_obj, attr=attr)
        return QuamDict(value, value_annotation=value_annotation)
    elif type(value) == list:
        value_annotation = _get_value_annotation(cls_or_obj=cls_or_obj, attr=attr)
        return QuamList(value, value_annotation=value_annotation)
    else:
        return value


def sort_quam_components(
    components: List["QuamComponent"], max_attempts=5
) -> List["QuamComponent"]:
    """Sort QuamComponent objects based on their config_settings.

    Args:
        components: A list of QuamComponent objects.
        max_attempts: The maximum number of attempts to sort the components.
            If the components aren't yet properly sorted after all these attempts,
            a warning is raised and the components are returned in the final attempted
            ordering.

    Returns:
        A sorted list of QuamComponent objects.

    Note:
        This function is used by
        [`QuamRoot.generate_config`][quam.core.quam_classes.QuamRoot.generate_config]
        to determine the order in which to add the components to the QUA config.
        This sorting isn't guaranteed to be successful.
    """
    sorted_components = components.copy()
    for _ in range(max_attempts):
        adjustments_made = False

        for component in components:
            if component.config_settings is None:
                continue

            component_idx = sorted_components.index(component)

            if "after" in component.config_settings:
                for after_component in component.config_settings["after"]:
                    after_component_idx = sorted_components.index(after_component)
                    if after_component_idx < component_idx:
                        continue
                    sorted_components.remove(after_component)
                    sorted_components.insert(component_idx, after_component)
                    adjustments_made = True

            if "before" in component.config_settings:
                for before_component in component.config_settings["before"]:
                    before_component_idx = sorted_components.index(before_component)
                    if before_component_idx > component_idx:
                        continue
                    sorted_components.remove(before_component)
                    sorted_components.insert(component_idx + 1, before_component)
                    adjustments_made = True

        if not adjustments_made:
            break
    else:
        warnings.warn(
            "Unable to sort QuamComponents based on config_settings. "
            "This may cause issues when generating the QUA config."
        )

    return sorted_components


def _quam_dataclass(cls=None, **kwargs):
    """Dataclass for QUAM classes.

    This class is used as a patch to maintain compatibility with Python 3.9, as
    these do not support the dataclass argument `kw_only`. This argument is needed to
    ensure inheritance of parent dataclasses is allowed.

    Args:
    - cls: The QUAM class to decorate.
    - kwargs: The arguments to pass to the dataclass decorator.
      By default, kw_only=True and eq=False are passed, though they can be overwritten.
    Notes:
    - This custom dataclass is no longer necessary once Python 3.9 support is dropped
    - The actual custom dataclass is `quam_dataclass` (without the underscore). This
      function is only used to trick type checkers into recognizing it as a dataclass.
    - From Python 3.10 onwards, this customized dataclass is no longer needed, as then
      the following two decorators are equivalent:
      - @quam_dataclass
      - @dataclass(eq=False, kw_only=True)
    """
    if cls is None:
        return partial(_quam_dataclass, **kwargs)

    kwargs.setdefault("kw_only", True)
    kwargs.setdefault("eq", False)

    if sys.version_info.minor > 9:
        return dataclass(cls, **kwargs)

    from quam.utils.dataclass import _quam_patched_dataclass

    return _quam_patched_dataclass(cls, **kwargs)


# Exec statement is needed to trick type checkers into recognizing it as a dataclass
# This will no longer be necessary once we drop support for Python 3.9
quam_dataclass = dataclass
exec("quam_dataclass = _quam_dataclass")


class ParentDescriptor:
    """Descriptor for the parent attribute of QuamBase.

    This descriptor is used to ensure that the parent attribute of a QuamBase
    object is not overwritten. This is to prevent the following situation:

    ```
    parent1 = QuamBase()
    parent2 = QuamBase()

    child = QuamBase()
    child.parent = parent1  # This is fine
    child.parent = parent2  # This raises an AttributeError
    ```
    """

    def __get__(self, instance, owner):
        if instance is None:
            return self

        if "parent" in instance.__dict__:
            return instance.__dict__["parent"]
        return None

    def __set__(self, instance, value):
        if value is None:
            instance.__dict__.pop("parent", None)
            return

        if "parent" in instance.__dict__ and instance.__dict__["parent"] is not value:
            cls = instance.__class__.__name__
            raise AttributeError(
                f"Cannot overwrite parent attribute of {cls}. "
                f"To modify {cls}.parent, first set {cls}.parent = None"
            )
        instance.__dict__["parent"] = value


class QuamBase(ReferenceClass):
    """Base class for any QUAM component class.

    args:
        parent: The parent of this object. This is automatically set when adding
            this object to another QuamBase object.
        config_settings: A dictionary of configuration settings for this object.
            This is used by [`QuamRoot.generate_config`][quam.core.quam_classes.QuamRoot.generate_config]
            to determine the order in which to add the components to the QUA config.
            Keys are "before" and "after", and the values are a list of QuamComponents

    Note:
        This class should not be used directly, but should generally be subclassed.
        The subclasses should be dataclasses.
    """

    parent: ClassVar["QuamBase"] = ParentDescriptor()
    _last_instantiated_root: ClassVar[Optional["QuamRoot"]] = None
    config_settings: ClassVar[Dict[str, Any]] = None
    _MAX_REFERENCE_DEPTH: ClassVar[int] = 10

    def __init__(self):
        # This prohibits instantiating without it being a dataclass
        # This means that we have to subclass this class and make it a dataclass
        if not is_dataclass(self):
            if type(self) in [QuamBase, QuamComponent, QuamRoot]:
                raise TypeError(
                    f"Cannot instantiate {self.__class__.__name__} directly. "
                    "Please create a subclass and make it a dataclass."
                )
            else:
                raise TypeError(
                    f"Cannot instantiate {self.__class__.__name__}. "
                    "Please make it a dataclass."
                )

    def _get_attr_names(self) -> List[str]:
        """Get names of all dataclass attributes of this object.

        Returns:
            List of attribute names.

        Raises:
            AssertionError if not a dataclass.
        """
        assert is_dataclass(self)
        return [data_field.name for data_field in fields(self)]

    def get_root(self) -> Optional[QuamRoot]:
        """Get the QuamRoot object of this object.

        This function recursively searches the parent chain for a QuamRoot object.
        If no QuamRoot object is found, it will return the last instantiated QuamRoot
        if it exists, else None.

        Returns:
            The root of this object, or None if no root is found.
        """
        if self.parent is not None:
            return self.parent.get_root()

        if self._last_instantiated_root is not None:
            warnings.warn(
                f"This component is not part of any QuamRoot, using last "
                f"instantiated QuamRoot. This is not recommended as it may lead to "
                f"unexpected behaviour. Component: {self.__class__.__name__}"
            )
            return self._last_instantiated_root
        return None

    def get_attr_name(self, attr_val: Any) -> str:
        """Get the name of an attribute that matches the value.

        Args:
            attr_val: The value of the attribute.

        Returns:
            The name of the attribute.

        Raises:
            AttributeError if not found.
        """
        for attr_name in self._get_attr_names():
            if self.get_raw_value(attr_name) is attr_val:
                return attr_name
        else:
            raise AttributeError(
                "Could not find name corresponding to attribute.\n"
                f"attribute: {attr_val}\n"
                f"obj: {self}"
            )

    def _attr_val_is_default(self, attr: str, val: Any) -> bool:
        """Check whether the value of an attribute is the default value.

        Args:
            attr: The name of the attribute.
            val: The value of the attribute.

        Returns:
            True if the value is the default value, False otherwise.
            False is also returned if the parent is not a dataclass
        """
        if not is_dataclass(self):
            return False

        dataclass_fields = fields(self)
        if not any(field.name == attr for field in dataclass_fields):
            return False

        field = next(field for field in dataclass_fields if field.name == attr)
        if field.default is not MISSING:
            return val == field.default
        elif field.default_factory is not MISSING:
            try:
                default_val = field.default_factory()
                return val == default_val
            except TypeError:
                return False

        return False

    @classmethod
    def _val_matches_attr_annotation(cls, attr: str, val: Any) -> bool:
        """Check whether the type of an attribute matches the annotation.

        The attribute type must exactly match the annotation.
        For dict and list, no additional type check of args is performed.
        """
        annotated_attrs = get_dataclass_attr_annotations(cls)
        if attr not in annotated_attrs["allowed"]:
            return False

        required_type = annotated_attrs["allowed"][attr]
        if type_is_optional(required_type):
            required_type = get_args(required_type)[0]

        if required_type == dict or get_origin(required_type) == dict:
            return isinstance(val, (dict, QuamDict))
        elif required_type == list or get_origin(required_type) == list:
            return isinstance(val, (list, QuamList))
        return type(val) == required_type

    def get_reference(
        self,
        attr: Optional[str] = None,
        relative_path: Optional[str] = None,
        follow_chain: bool = False,
    ) -> Optional[str]:
        """Get the reference path of this object or one of its attributes.

        Args:
            attr: The optional attribute to get the reference path for.
                If None, the reference path of the object itself is returned.
            relative_path: The optional relative path to join with the reference path.
            follow_chain: If True and attr is a reference, follow the reference chain
                to return the ultimate target reference. Default is False for backward
                compatibility. Only applies when attr is specified.

        Returns:
            The reference path of this object or the specified attribute.

        Raises:
            ValueError: If both attr and relative_path are specified, or if follow_chain
                is True but attr is not a reference.

        Examples:
            We assume a QuamRoot object with a component "elem".
            - elem.get_reference() == "#/elem"
            - elem.get_reference(attr="child") == "#/elem/child"
            - elem.get_reference(relative_path="#./child") == "#/elem/child"
            - elem.get_reference(relative_path="#../child") == "#/child"
            - elem.get_reference(relative_path="#./child/grandchild") == "#/elem/child/grandchild"

            With follow_chain=True (if attr contains a reference to another reference):
            - elem.get_reference(attr="chain_ref", follow_chain=True)  # returns ultimate target
        """

        if attr is not None and relative_path is not None:
            raise ValueError(
                "Cannot specify both attr and relative_path. "
                "Please specify only one of them."
            )

        # Handle follow_chain before computing the reference path
        if follow_chain and attr is not None:
            raw_value = self.get_raw_value(attr)
            if not string_reference.is_reference(raw_value):
                raise ValueError(
                    f"Cannot follow reference chain for attr '{attr}': "
                    f"value is not a reference (value={raw_value})"
                )
            # Follow the reference chain to get the ultimate target
            try:
                target_obj, target_attr = self._follow_reference_chain(self, attr)
                # Return the reference path to the ultimate target
                return target_obj.get_reference(attr=target_attr)
            except (AttributeError, ValueError, RecursionError) as e:
                raise ValueError(
                    f"Failed to follow reference chain for attr '{attr}': {e}"
                ) from e

        if self.parent is None:
            raise AttributeError(
                "Unable to extract reference path for {self}: No parent defined"
            )

        try:
            base_reference = self.parent.get_reference()
            if base_reference == "#/":
                base_reference = "#"
        except AttributeError:
            raise AttributeError(
                f"Unable to extract reference path for {self}: Could not get "
                f"reference path for parent {self.parent}"
            )

        try:
            attr_name = self.parent.get_attr_name(self)
        except AttributeError:
            raise AttributeError(
                f"Unable to extract reference path for {self}: Could not get "
                f"attribute name from parent {self.parent}"
            )

        reference = f"{base_reference}/{attr_name}"

        if relative_path is not None:
            if not string_reference.is_reference(relative_path):
                raise ValueError(
                    f"Unable to extract reference path for {self}: "
                    f"relative_path {relative_path} is not a reference"
                )

            reference = string_reference.join_references(reference, relative_path)
        elif attr is not None:
            reference = f"{reference}/{attr}"

        return reference

    def get_attrs(
        self, follow_references: bool = False, include_defaults: bool = True
    ) -> Dict[str, Any]:
        """Get all attributes and corresponding values of this object.

        Args:
            follow_references: Whether to follow references when getting the value.
                If False, the reference will be returned as a string.
            include_defaults: Whether to include attributes that have the default
                value.

        Returns:
            A dictionary of attribute names and values.

        """
        attr_names = self._get_attr_names()

        skip_attrs = getattr(self, "_skip_attrs", [])
        attr_names = [attr for attr in attr_names if attr not in skip_attrs]

        # Apply skip_save metadata filter
        if is_dataclass(self):
            skip_save_attrs = [
                field.name
                for field in fields(self)
                if field.metadata.get("skip_save", False)
            ]
            attr_names = [attr for attr in attr_names if attr not in skip_save_attrs]

        if not follow_references:
            attrs = {attr: self.get_raw_value(attr) for attr in attr_names}
        else:
            attrs = {attr: getattr(self, attr) for attr in attr_names}

        if not include_defaults:
            attrs = {
                attr: val
                for attr, val in attrs.items()
                if not self._attr_val_is_default(attr, val)
            }
        return attrs

    def to_dict(
        self, follow_references: bool = False, include_defaults: bool = True
    ) -> Dict[str, Any]:
        """Convert this object to a dictionary.

        Args:
            follow_references: Whether to follow references when getting the value.
                If False, the reference will be returned as a string.
            include_defaults: Whether to include attributes that have the default value.

        Returns:
            A dictionary representation of this object.
            Any QuamBase objects will be recursively converted to dictionaries.

        Note:
            If the value of an attribute does not match the annotation, the
            `"__class__"` key will be added to the dictionary. This is to ensure
            that the object can be reconstructed when loading from a file.
        """
        attrs = self.get_attrs(
            follow_references=follow_references, include_defaults=include_defaults
        )
        quam_dict = {}
        for attr, val in attrs.items():
            if isinstance(val, QuamBase):
                quam_dict[attr] = val.to_dict(
                    follow_references=follow_references,
                    include_defaults=include_defaults,
                )
            else:
                quam_dict[attr] = val

        # Add __class__ to specify the class of the object
        quam_dict["__class__"] = get_full_class_path(self)

        return quam_dict

    def iterate_components(
        self, skip_elems: Optional[Sequence["QuamBase"]] = None
    ) -> Generator["QuamBase", None, None]:
        """Iterate over all QuamBase objects in this object, including nested objects.

        Args:
            skip_elems: A sequence of QuamBase objects to skip.
                This is used to prevent infinite loops when iterating over nested
                objects.

        Returns:
            A generator of QuamBase objects.
        """
        if skip_elems is None:
            skip_elems = []

        # We don't use "self in skip_elems" because we want to check for identity,
        # not equality. The reason is that you would otherwise have to instantiate
        # dataclasses using @dataclass(eq=False)
        in_skip_elems = any(self is elem for elem in skip_elems)
        if isinstance(self, QuamComponent) and not in_skip_elems:
            skip_elems.append(self)
            yield self

        attrs = self.get_attrs(follow_references=False, include_defaults=True)

        for attr_val in attrs.values():
            if any(attr_val is elem for elem in skip_elems):
                continue

            if isinstance(attr_val, QuamBase):
                yield from attr_val.iterate_components(skip_elems=skip_elems)

    def _is_reference(self, attr: str) -> bool:
        """Check whether an attribute is a reference.

        Args:
            attr: The name of the attribute.

        Returns:
            True if the attribute is a reference, False otherwise.

        Note:
            This function is used from the ReferenceClass class.
        """
        return string_reference.is_reference(attr)

    def _get_referenced_value(self, reference: str) -> Any:
        """Get the value of an attribute by reference

        Args:
            reference: The reference to the attribute.

        Returns:
            The value of the attribute, or the reference if it is not a reference

        Note:
            This function is used from the ReferenceClass class.
        """
        if not string_reference.is_reference(reference):
            return reference

        root = self.get_root()
        if string_reference.is_absolute_reference(reference) and root is None:
            warnings.warn(
                f"No QuamRoot initialized, cannot retrieve absolute reference "
                f"{reference} from {self.__class__.__name__}"
            )
            return reference

        try:
            return string_reference.get_referenced_value(self, reference, root=root)
        except InvalidReferenceError as e:
            try:
                ref = f"{self.__class__.__name__}: {self.get_reference()}"
            except Exception:
                ref = self.__class__.__name__

            # Construct message using the chained exception's message
            base_error_msg = str(e)
            component_ref_str = f"component {ref}"
            msg = (
                f'Could not get reference "{reference}" from {component_ref_str}. '
                f"Error: {base_error_msg}"
            )

            try:
                cfg = get_quam_config()
                raise_error_missing_reference = cfg.raise_error_missing_reference
            except FileNotFoundError:
                # Default behavior if config file not found
                raise_error_missing_reference = False

            if not raise_error_missing_reference:
                warnings.warn(msg)
            else:
                # Re-raise with context, keeping the original exception chain
                raise InvalidReferenceError(msg) from e
            return reference

    def print_summary(self, indent: int = 0):
        """Print a summary of the QuamBase object.

        Args:
            indent: The number of spaces to indent the summary.
        """
        if self.get_root() is self:
            full_name = "QUAM:"
        elif self.parent is None:
            full_name = f"{self.__class__.__name__} (parent unknown):"
        else:
            try:
                attr_name = self.parent.get_attr_name(self)
                full_name = f"{attr_name}: {self.__class__.__name__}"
            except AttributeError:
                full_name = f"{self.__class__.__name__}:"

        if not self.get_attrs():
            print(" " * indent + f"{full_name} Empty")
            return

        print(" " * indent + f"{full_name}")
        for attr, val in self.get_attrs().items():
            if isinstance(val, str):
                val = f'"{val}"'
            if isinstance(val, QuamBase):
                val.print_summary(indent=indent + 2)
            else:
                print(" " * (indent + 2) + f"{attr}: {val}")

    def _follow_reference_chain(
        self, obj: "QuamBase", attr: str, max_depth: Optional[int] = None
    ) -> tuple["QuamBase", str]:
        """Recursively follow a reference chain to find the ultimate target.

        This method follows reference strings through nested references until
        reaching a non-reference value. It is used internally by
        set_at_reference() and get_reference() to handle chained references.

        Args:
            obj: The QuamBase object containing the attribute.
            attr: The attribute name to follow (may be a list index like "0").
            max_depth: Maximum recursion depth to prevent infinite loops
                (default: QuamBase._MAX_REFERENCE_DEPTH).

        Returns:
            A tuple of (target_object, final_attr_name) where:
            - target_object: The QuamBase object containing the ultimate value
            - final_attr_name: The final attribute name (after following all chains)

        Raises:
            AttributeError: If a reference in the chain cannot be resolved.
            RecursionError: If max_depth is exceeded (indicates circular reference).
        """
        # Normalize attr to string for consistent handling
        attr = str(attr)

        if max_depth is None:
            max_depth = self._MAX_REFERENCE_DEPTH
        if max_depth <= 0:
            raise RecursionError(
                f"Reference chain exceeded maximum depth of {self._MAX_REFERENCE_DEPTH}. "
                f"Possible circular reference starting from {obj.get_attr_path()}"
            )

        # Handle list/dict index access specially
        if attr.isdigit() and isinstance(obj, (list, UserList, QuamList)):
            # For list indices, get the raw element directly from the list
            index = int(attr)
            if isinstance(obj, QuamList):
                raw_value = obj.data[index]
            else:
                raw_value = obj[index]
        else:
            raw_value = obj.get_raw_value(attr)

        # Base case: value is not a reference
        if not string_reference.is_reference(raw_value):
            return obj, attr

        # Recursive case: follow the reference chain
        parent_reference, parent_attr = string_reference.split_reference(raw_value)
        if not parent_attr:
            raise ValueError(
                f"Invalid reference {raw_value}: must have an attribute part"
            )

        parent_obj = obj._get_referenced_value(parent_reference)
        if not isinstance(parent_obj, QuamBase):
            # This can happen when:
            # 1. Broken reference: _get_referenced_value returns the reference string
            # 2. Reference to a non-QuamBase value (e.g., primitive)
            # Raise AttributeError to match the expected contract for set_at_reference
            raise AttributeError(
                f"Cannot follow reference chain: '{parent_reference}' resolved to "
                f"{type(parent_obj).__name__}, not a QuamBase object"
            )

        # Recursively follow the chain
        return self._follow_reference_chain(parent_obj, parent_attr, max_depth - 1)

    def set_at_reference(
        self, attr: str, value: Any, allow_non_reference: bool = True
    ):
        """Follow the reference of an attribute and set the value at the reference.

        This method follows reference chains recursively. If an attribute contains
        a reference to another reference, both references are preserved while the
        ultimate target value is updated.

        Args:
            attr: The attribute to set the value at the reference of.
            value: The value to set.
            allow_non_reference: Whether to allow the attribute to be a non-reference.
                If True (default), non-reference attributes are allowed. If False,
                the attribute must be a reference or an error is raised.

        Raises:
            ValueError: If the attribute is not a reference and `allow_non_reference` is
                False.
            ValueError: If the reference is invalid, e.g. "#./" since it has no
                attribute.
        """
        raw_value = self.get_raw_value(attr)
        if not string_reference.is_reference(raw_value):
            if not allow_non_reference:
                raise ValueError(
                    f"Cannot set at reference because attr '{attr}' is not a reference. "
                    f"'{attr}' = {raw_value}"
                )
            target_obj, target_attr = self, attr
        else:
            # Follow the reference chain to get the ultimate target
            target_obj, target_attr = self._follow_reference_chain(self, attr)

        # Use __setitem__ for dict/list types, otherwise use setattr
        if isinstance(target_obj, (dict, list, UserDict, UserList, QuamDict, QuamList)):
            # Convert string index to int for list types
            if (
                isinstance(target_obj, (list, UserList, QuamList))
                and target_attr.isdigit()
            ):
                target_attr = int(target_attr)
            target_obj[target_attr] = value
        else:
            setattr(target_obj, target_attr, value)


# Type annotation for QuamRoot, can be replaced by typing.Self from Python 3.11
QuamRootType = TypeVar("QuamRootType", bound="QuamRoot")


class QuamRoot(QuamBase):
    """Base class for the root of a QUAM object.

    This class should be subclassed and made a dataclass.

    Note:
        This class should not be used directly, but should generally be subclassed and
        made a dataclass. The dataclass fields should correspond to the QUAM root
        structure.
    """

    def __post_init__(self):
        QuamBase._last_instantiated_root = self
        self.serialiser = self.get_serialiser()
        super().__post_init__()

    def __setattr__(self, name, value):
        converted_val = convert_dict_and_list(value, cls_or_obj=self, attr=name)
        super().__setattr__(name, converted_val)

        if isinstance(converted_val, QuamBase) and name != "parent":
            converted_val.parent = self

    @classmethod
    def get_serialiser(cls) -> AbstractSerialiser:
        """Get the serialiser for the QuamRoot class, which is the JSONSerialiser.

        This method can be overridden by subclasses to provide a custom serialiser.
        """
        return JSONSerialiser()

    def get_reference(
        self, attr: Optional[str] = None, relative_path: Optional[str] = None
    ) -> Optional[str]:
        if attr is not None:
            return f"#/{attr}"

        reference = "#/"
        if relative_path is not None:
            reference = string_reference.join_references(reference, relative_path)
        return reference

    def get_root(self: QuamRootType) -> QuamRootType:
        """Get the QuamRoot object of this object, i.e. the object itself.

        This QuamRoot function overrides the QuamBase function to return the object
        itself, rather than following the parent chain.

        Returns:
            The current QuamRoot object (self).
        """
        return self

    def save(
        self,
        path: Optional[Union[Path, str]] = None,
        content_mapping: Optional[Dict[str, str]] = None,
        include_defaults: Optional[bool] = None,
        ignore: Optional[Sequence[str]] = None,
    ):
        """Save the entire QuamRoot object to a file. This includes nested objects.

        Args:
            path: The path to save the file to. If None, the path will be extracted from
                the `state_path` attribute of the serialiser, which could be set by the
                quam config file or environment variable.
            content_mapping: Optional mapping of component names to filenames. This can
                be used to save different parts of the QuamRoot object to different
                files.
            include_defaults: Whether to include attributes that have the default
                value.
            ignore: A list of components to ignore.
        """
        self.serialiser.save(
            quam_obj=self,
            path=path,
            content_mapping=content_mapping,
            include_defaults=include_defaults,
            ignore=ignore,
        )

    @classmethod
    def load(
        cls: type[QuamRootType],
        filepath_or_dict: Optional[Union[str, Path, dict]] = None,
        validate_type: bool = True,
        fix_attrs: bool = True,
    ) -> QuamRootType:
        """Load a QuamRoot object from a file.

        Args:
            filepath_or_dict: The path to the file/folder to load, or a dictionary.
                The dictionary would be the result from a call to `QuamRoot.save()`
                Can be omitted, in which case the serialiser will use the default state
                path, which is typically defined in the quam config file.
            validate_type: Whether to validate the type of all attributes while loading.
            fix_attrs: Whether attributes can be added to QuamBase objects that are not
                defined as dataclass fields.

        Returns:
            A QuamRoot object instantiated from the file/folder/dict.
        """
        if isinstance(filepath_or_dict, dict):
            contents = filepath_or_dict
        else:
            serialiser = cls.get_serialiser()
            contents, _ = serialiser.load(filepath_or_dict)

        return instantiate_quam_class(
            quam_class=cls,
            contents=contents,
            fix_attrs=fix_attrs,
            validate_type=validate_type,
        )

    def generate_config(self) -> DictQuaConfig:
        """Generate the QUA configuration from the QUAM object.

        Returns:
            A dictionary with the QUA configuration.

        Note:
            This function collects all the nested QuamComponent objects and calls
            `QuamComponent.apply_to_config` on them.
        """
        qua_config = deepcopy(qua_config_template)

        quam_components = list(self.iterate_components())
        sorted_components = sort_quam_components(quam_components)

        for quam_component in sorted_components:
            quam_component.apply_to_config(qua_config)

        generate_config_final_actions(qua_config)

        return qua_config


class QuamComponent(QuamBase):
    """Base class for any QUAM component class.

    Examples of QuamComponent classes are [`Mixer`][quam.components.hardware.Mixer],
    [`LocalOscillator`][quam.components.hardware.LocalOscillator],
    [`Pulse`][quam.components.pulses.Pulse], etc.

    Note:
        This class should be subclassed and made a dataclass.
    """

    def __setattr__(self, name, value):
        converted_val = convert_dict_and_list(value, cls_or_obj=self, attr=name)
        super().__setattr__(name, converted_val)

        if isinstance(converted_val, QuamBase) and name != "parent":
            converted_val.parent = self

    def apply_to_config(self, config: dict) -> None:
        """Add information to the QUA configuration, such as pulses and waveforms.

        Args:
            config: The QUA configuration dictionary. Initially this is a nearly empty
                dictionary, but

        Note:
            This function is called by
            [`QuamRoot.generate_config`][quam.core.quam_classes.QuamRoot.generate_config].

        Note:
            The config has a starting template, defined at [`quam.core.qua_config_template`][]
        """
        ...


@quam_dataclass
class QuamDict(UserDict, QuamBase):
    """A QUAM dictionary class.

    Any dict added to a `QuamBase` object is automatically converted to a `QuamDict`.
    The `QuamDict` adds the following functionalities to a dict:
    - Values can be references (see below)
    - Keys can also be accessed through attributes (e.g. `d.a` instead of `d["a"]`)

    # QuamDict references
    QuamDict values can be references, which are strings that start with `#`. See the
    documentation for details on references. An example is shown here:
    ```
    d = QuamDict({"a": 1, "b": "#./a"})
    assert d["b"] == 1
    ```

    Warning:
        This class is a subclass of `QuamBase`, but also of `UserDict`. As a result,
        it can be used as a normal dictionary, but it is not a subclass of `dict`.
    """

    _value_annotation: ClassVar[type] = None

    def __init__(self, dict=None, /, value_annotation: type = None, **kwargs):
        self.__dict__["data"] = {}
        self.__dict__["_value_annotation"] = value_annotation
        self.__dict__["_initialized"] = True
        super().__init__(dict, **kwargs)

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError as e:
            try:
                repr = f"{self.__class__.__name__}: {self.get_reference()}"
            except Exception:
                repr = self.__class__.__name__
            raise AttributeError(f'{repr} has no attribute "{key}"') from e

    def __setattr__(self, key, value):
        if key in ["data", "parent", "config_settings", "_initialized"]:
            super().__setattr__(key, value)
        else:
            self[key] = value

    def __getitem__(self, i):
        elem = super().__getitem__(i)

        if not string_reference.is_reference(elem):
            return elem

        try:
            target_obj, target_attr = self._follow_reference_chain(self, i)
            # Handle list/dict indices that result from following the chain
            if target_attr.isdigit() and isinstance(
                target_obj, (list, UserList, QuamList)
            ):
                return target_obj[int(target_attr)]
            elif isinstance(target_obj, (dict, UserDict, QuamDict)):
                return target_obj[target_attr]
            else:
                return target_obj.get_raw_value(target_attr)
        except (AttributeError, KeyError, ValueError):
            # Chain couldn't be followed (broken reference, missing attribute, etc.)
            # Return the reference string - _get_referenced_value handles warnings
            return self._get_referenced_value(elem)

    # Overriding methods from UserDict
    def __setitem__(self, key, value):
        value = convert_dict_and_list(value)
        self._is_valid_setattr(key, value, error_on_False=True)
        super().__setitem__(key, value)

        if isinstance(value, QuamBase):
            value.parent = self

    def __eq__(self, other) -> bool:
        if isinstance(other, dict):
            return self.data == other
        return super().__eq__(other)

    def __repr__(self) -> str:
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message="^Could not get reference*",
                category=UserWarning,
            )
            return super().__repr__()

    # QUAM methods
    def _get_attr_names(self):
        return list(self.data.keys())

    def get_attrs(
        self, follow_references=False, include_defaults=True
    ) -> Dict[str, Any]:
        # TODO implement reference kwargs
        return self.data

    def get_attr_name(self, attr_val: Any) -> Union[str, int]:
        """Get the name of an attribute that matches the value.

        Args:
            attr_val: The value of the attribute.

        Returns:
            The name of the attribute. This can also be an int depending on the dict key

        Raises:
            AttributeError if not found.
        """
        for attr_name in self._get_attr_names():
            if attr_name in self and self.get_raw_value(attr_name) is attr_val:
                return attr_name
        else:
            raise AttributeError(
                "Could not find name corresponding to attribute.\n"
                f"attribute: {attr_val}\n"
                f"obj: {self}"
            )

    def _val_matches_attr_annotation(self, attr: str, val: Any) -> bool:
        """Check whether the type of an attribute matches the annotation.

        Called by [`QuamDict.to_dict`][quam.core.quam_classes.QuamDict.to_dict] to
        determine whether to add the __class__ key.

        Args:
            attr: The name of the attribute. Unused but added to match parent signature
            val: The value of the attribute.

        Note:
            The attribute val is compared to `QuamDict._value_annotation`, which is set
            when a dict is converted to a `QuamDict` using `convert_dict_and_list`.
        """
        if isinstance(val, (QuamDict, QuamList)):
            return True
        if self._value_annotation is None:
            return False
        return type(val) == self._value_annotation

    def _attr_val_is_default(self, attr: str, val: Any):
        """Check whether the value of an attribute is the default value.

        Overrides parent method.
        Since a QuamDict does not have any fixed attrs, this is always False.

        """
        return False

    def get_raw_value(self, attr: str) -> Any:
        """Get the value of an attribute without following references.

        Args:
            attr: The name of the attribute.

        Returns:
            The value of the attribute. If the value is a reference, it returns the
            reference string instead of the value it is referencing.
        """
        try:
            return self.__dict__["data"][attr]
        except KeyError as e:
            raise AttributeError(
                "Cannot get raw (unreferenced)value from attribute {attr} that does not"
                " exist in {self}"
            ) from e

    def iterate_components(
        self, skip_elems: Sequence[QuamBase] = None
    ) -> Generator["QuamBase", None, None]:
        """Iterate over all QuamBase objects in this object, including nested objects.

        Args:
            skip_elems: A sequence of QuamBase objects to skip.
                This is used to prevent infinite loops when iterating over nested
                objects.

        Returns:
            A generator of QuamBase objects.
        """
        if skip_elems is None:
            skip_elems = []

        for attr_val in self.data.values():
            if any(attr_val is elem for elem in skip_elems):
                continue

            if isinstance(attr_val, QuamBase):
                yield from attr_val.iterate_components(skip_elems=skip_elems)

    def to_dict(
        self, follow_references: bool = False, include_defaults: bool = True
    ) -> dict:
        """Convert this object to a dictionary.

        Ensures all child QUAM objects are also converted to dictionaries.

        Args:
            follow_references: Whether to follow references when getting the value.
                If False, the reference will be returned as a string.
            include_defaults: Whether to include attributes that have the default
                value.

        Returns:
            A dictionary representation of the object.
        """
        quam_dict = super().to_dict(
            follow_references=follow_references,
            include_defaults=include_defaults,
        )

        # Remove __class__ from the dictionary as it's the default for a dict
        quam_dict.pop("__class__", None)

        return quam_dict


@quam_dataclass
class QuamList(UserList, QuamBase):
    """A QUAM list class.

    Any list added to a `QuamBase` object is automatically converted to a `QuamList`.
    The `QuamList` adds the following functionalities to a list:
    - Elements can be references (see below)

    # QuamList references
    QuamList values can be references, which are strings that start with `#`. See the
    documentation for details on references. An example is shown here:
    ```
    d = QuamList([1, "#./0"]])
    assert d[1] == 1
    ```

    Warning:
        This class is a subclass of `QuamBase`, but also of `UserList`. As a result,
        it can be used as a normal list, but it is not a subclass of `list`.
    """

    _value_annotation: ClassVar[type] = None

    def __init__(self, *args, value_annotation: type = None):
        self._value_annotation = value_annotation

        # We manually add elements using extend instead of passing to super()
        # To ensure that any dicts and lists get converted to QuamDict and QuamList
        super().__init__()
        if args:
            self.extend(*args)

    # Overloading methods from UserList
    def __eq__(self, value: object) -> bool:
        return super().__eq__(value)

    def __repr__(self) -> str:
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message="^Could not get reference*",
                category=UserWarning,
            )
            return super().__repr__()

    def __getitem__(self, i):
        elem = super().__getitem__(i)
        if isinstance(i, slice):
            if isinstance(elem, QuamList):
                elem.parent = self.parent
            # This automatically gets the referenced values
            return list(elem)

        elif not string_reference.is_reference(elem):
            return elem

        try:
            target_obj, target_attr = self._follow_reference_chain(self, i)
            # Handle list/dict indices that result from following the chain
            if target_attr.isdigit() and isinstance(
                target_obj, (list, UserList, QuamList)
            ):
                return target_obj[int(target_attr)]
            elif isinstance(target_obj, (dict, UserDict, QuamDict)):
                return target_obj[target_attr]
            else:
                return target_obj.get_raw_value(target_attr)
        except (AttributeError, KeyError, ValueError):
            # Chain couldn't be followed (broken reference, missing attribute, etc.)
            # Return the reference string - _get_referenced_value handles warnings
            return self._get_referenced_value(elem)

    def __setitem__(self, i, item):
        converted_item = convert_dict_and_list(item)
        super().__setitem__(i, converted_item)

        if isinstance(converted_item, QuamBase):
            converted_item.parent = self

    def __iadd__(self, other: Iterable):
        converted_other = [convert_dict_and_list(elem) for elem in other]
        return super().__iadd__(converted_other)

    def append(self, item: Any) -> None:
        converted_item = convert_dict_and_list(item)

        if isinstance(converted_item, QuamBase):
            converted_item.parent = self

        return super().append(converted_item)

    def insert(self, i: int, item: Any) -> None:
        converted_item = convert_dict_and_list(item)

        if isinstance(converted_item, QuamBase):
            converted_item.parent = self

        return super().insert(i, converted_item)

    def extend(self, iterable: Iterator) -> None:
        converted_iterable = [convert_dict_and_list(elem) for elem in iterable]
        for converted_item in converted_iterable:
            if isinstance(converted_item, QuamBase):
                converted_item.parent = self

        return super().extend(converted_iterable)

    # Quam methods
    def _val_matches_attr_annotation(self, attr: str, val: Any) -> bool:
        """Check whether the type of an attribute matches the annotation.

        Called by QuamList.to_dict to determine whether to add the __class__ key.
        For the QuamList, we compare the type to the _value_annotation.
        """
        if isinstance(val, (QuamDict, QuamList)):
            return True
        if self._value_annotation is None:
            return False
        return type(val) == self._value_annotation

    def get_attr_name(self, attr_val: Any) -> str:
        for k, elem in enumerate(self.data):
            if elem is attr_val:
                return str(k)
        else:
            raise AttributeError(
                "Could not find name corresponding to attribute"
                f"attribute: {attr_val}\n"
                f"obj: {self}"
            )

    def to_dict(
        self, follow_references: bool = False, include_defaults: bool = True
    ) -> list:
        """Convert this object to a list, usually as part of a dictionary representation.

        Args:
            follow_references: Whether to follow references when getting the value.
                If False, the reference will be returned as a string.
            include_defaults: Whether to include attributes that have the default
                value.

        Returns:
            A list with the values of this object. Any QuamBase objects will be
            recursively converted to dictionaries.

        Note:
            If the value of an attribute does not match the annotation of
            `QuamList._value_annotation`, the `"__class__"` key will be added to the
            dictionary. This is to ensure that the object can be reconstructed when
            loading from a file.
        """
        quam_list = []
        for val in self.data:
            if isinstance(val, QuamBase):
                quam_list.append(
                    val.to_dict(
                        follow_references=follow_references,
                        include_defaults=include_defaults,
                    )
                )
            else:
                quam_list.append(val)
        return quam_list

    def iterate_components(
        self, skip_elems: List[QuamBase] = None
    ) -> Generator["QuamBase", None, None]:
        """Iterate over all QuamBase objects in this object, including nested objects.

        Args:
            skip_elems: A list of QuamBase objects to skip.
                This is used to prevent infinite loops when iterating over nested
                objects.

        Returns:
            A generator of QuamBase objects.
        """
        if skip_elems is None:
            skip_elems = []

        for attr_val in self.data:
            if any(attr_val is elem for elem in skip_elems):
                continue

            if isinstance(attr_val, QuamBase):
                yield from attr_val.iterate_components(skip_elems=skip_elems)

    def get_attrs(
        self, follow_references: bool = False, include_defaults: bool = True
    ) -> Dict[str, Any]:
        raise NotImplementedError("QuamList does not have attributes")

    def print_summary(self, indent: int = 0):
        """Print a summary of the QuamBase object.

        Args:
            indent: The number of spaces to indent the summary.
        """

        if self.parent is None:
            full_name = f"{self.__class__.__name__} (parent unknown):"
        else:
            try:
                attr_name = self.parent.get_attr_name(self)
                full_name = f"{attr_name}: {self.__class__.__name__}"
            except AttributeError:
                full_name = f"{self.__class__.__name__}:"

        if not self.data:
            print(" " * indent + f"{full_name} = []")
            return

        print(" " * indent + f"{full_name}:")
        if len(str(self.data)) + 2 * indent < 80:
            print(" " * (indent + 2) + f"{self.data}")
        else:
            for k, val in enumerate(self.data):
                if isinstance(val, QuamBase):
                    val.print_summary(indent=indent + 2)
                else:
                    print(" " * (indent + 2) + f"{k}: {val}")



import pytest
import json
import sys
import os
from pathlib import Path
from typing import Dict, Any, Union, Optional
from dataclasses import field

# Ensure quam imports work correctly relative to the project structure
# This might need adjustment based on your exact test setup location
# import sys
# sys.path.insert(0, str(Path(__file__).parent.parent)) # Assuming tests are one level above quam package

from quam.core import QuamRoot, QuamComponent, quam_dataclass
from quam.serialisation.json import JSONSerialiser, convert_int_keys

# Import the actual config model to get the expected version
try:
    from quam.config.models import QuamConfig

    EXPECTED_CONFIG_VERSION = QuamConfig.version
except ImportError:
    # Fallback if models can't be imported (less robust)
    EXPECTED_CONFIG_VERSION = 1


# Mock QUAM Components


@quam_dataclass
class MockChildComponent(QuamComponent):
    """A simple nested component."""

    id: Union[str, int] = "child_id"
    child_value: int = 50  # Default value


@quam_dataclass
class MockMainComponent(QuamComponent):
    """A mock component that can hold a child."""

    id: Union[str, int] = "main_id"
    value: int = 1  # Default value
    child: Optional[MockChildComponent] = None


@quam_dataclass
class MockQuamRoot(QuamRoot):
    """A mock QuamRoot for testing serialization."""

    # Use default_factory for mutable defaults like dict
    components: Dict[str, MockMainComponent] = field(default_factory=dict)
    wiring: Dict[str, Any] = field(default_factory=dict)
    other: str = "default_other"
    default_val: int = 10  # Example default


# Note: prevent_default_config_loading fixture is now in tests/conftest.py
# and applies to all tests automatically


@pytest.fixture
def serialiser():
    """Basic JSONSerialiser instance."""
    return JSONSerialiser()


@pytest.fixture
def serialiser_with_defaults():
    """JSONSerialiser instance configured to include defaults."""
    return JSONSerialiser(include_defaults=True)


@pytest.fixture
def sample_quam_object():
    """Provides a simple mock QuamRoot object for testing."""
    root = MockQuamRoot(
        components={
            "comp1": MockMainComponent(id="c1"),  # Default value=1
            "comp2": MockMainComponent(id="c2", value=5),  # Non-default value
        },
        wiring={"w1": "p1", "w2": "p2"},
        other="specific_value",  # Non-default value
        # default_val uses its default (10)
    )
    return root


@pytest.fixture
def sample_quam_object_nested():
    """Provides a mock QuamRoot object with nested components."""
    child_comp = MockChildComponent(id="child1", child_value=99)  # Non-default
    parent_comp = MockMainComponent(id="parent1", child=child_comp)

    root = MockQuamRoot(
        components={"parent": parent_comp},
        wiring={"w1": "p1"},
        default_val=15,  # Non-default
    )
    return root


@pytest.fixture
def sample_dict_basic():
    """A simple dictionary for basic load/save tests."""
    return {"a": 1, "b": "hello", "c": [1, 2, 3]}


@pytest.fixture
def sample_dict_int_keys():
    """Represents how a dict with int keys is stored in JSON (string keys)."""
    return {"1": "one", "5": "five", "key": "value"}


@pytest.fixture
def setup_single_file(tmp_path, sample_dict_basic):
    """Creates a single JSON file for load testing."""
    filepath = tmp_path / "load_single.json"
    with filepath.open("w", encoding="utf-8") as f:
        json.dump(sample_dict_basic, f)
    return filepath


@pytest.fixture
def setup_directory_basic(tmp_path, sample_dict_basic):
    """Creates a directory with a single default JSON file."""
    dirpath = tmp_path / "load_dir_basic"
    dirpath.mkdir()
    filepath = dirpath / JSONSerialiser.default_filename  # state.json
    with filepath.open("w", encoding="utf-8") as f:
        json.dump(sample_dict_basic, f)
    return dirpath


@pytest.fixture
def setup_directory_split(tmp_path):
    """
    Creates a directory with split JSON files based on sample_quam_object structure.
    """
    dirpath = tmp_path / "load_dir_split"
    dirpath.mkdir()
    # Content based on sample_quam_object (excluding defaults)
    wiring_content = {"wiring": {"w1": "p1", "w2": "p2"}}
    components_content = {
        "components": {"comp1": {"id": "c1"}, "comp2": {"id": "c2", "value": 5}}
    }
    default_content = {"other": "specific_value"}

    with (dirpath / "wiring.json").open("w", encoding="utf-8") as f:
        json.dump(wiring_content, f)
    with (dirpath / "components.json").open("w", encoding="utf-8") as f:
        json.dump(components_content, f)
    with (dirpath / JSONSerialiser.default_filename).open("w", encoding="utf-8") as f:
        json.dump(default_content, f)
    # We also need the __class__ attribute for proper loading of the root
    default_content_root_class = default_content.copy()
    # Determine class path dynamically
    root_class_path = f"{MockQuamRoot.__module__}.{MockQuamRoot.__name__}"
    default_content_root_class["__class__"] = root_class_path
    with (dirpath / JSONSerialiser.default_filename).open("w", encoding="utf-8") as f:
        json.dump(default_content_root_class, f)

    # Also save __class__ for components if needed by loader (though QuamRoot.load
    # handles this)
    components_content_class = components_content.copy()
    comp_class_path = f"{MockMainComponent.__module__}.{MockMainComponent.__name__}"
    for comp_data in components_content_class["components"].values():
        comp_data["__class__"] = comp_class_path
    with (dirpath / "components.json").open("w", encoding="utf-8") as f:
        json.dump(components_content_class, f)

    return dirpath


@pytest.fixture
def setup_directory_conflict(tmp_path):
    """Creates a directory with files having conflicting keys."""
    dirpath = tmp_path / "load_dir_conflict"
    dirpath.mkdir()
    file1_content = {"a": 1, "b": "original"}
    file2_content = {"b": "overwrite", "c": 3}  # 'b' conflicts

    with (dirpath / "file1.json").open("w", encoding="utf-8") as f:
        json.dump(file1_content, f)
    with (dirpath / "file2.json").open("w", encoding="utf-8") as f:
        json.dump(file2_content, f)
    return dirpath


@pytest.fixture
def setup_directory_int_keys(tmp_path, sample_dict_int_keys):
    """Creates a directory with integer keys saved as strings."""
    dirpath = tmp_path / "load_dir_int_keys"
    dirpath.mkdir()
    filepath = dirpath / "int_keys.json"
    with filepath.open("w", encoding="utf-8") as f:
        json.dump(sample_dict_int_keys, f)
    return dirpath


# Mocking fixtures
@pytest.fixture
def mock_env(monkeypatch):
    """Fixture to mock environment variables."""
    original_env = os.environ.copy()

    def set_env(var, value):
        if value is None:
            monkeypatch.delenv(var, raising=False)
        else:
            monkeypatch.setenv(var, str(value))

    yield set_env
    # Restore original environment state after test
    os.environ.clear()
    os.environ.update(original_env)


# UPDATED mock_config fixture
@pytest.fixture
def mock_config(monkeypatch):
    """Fixture to mock quam.config get_quam_config function."""

    class MockConfigContainer:
        """Helper class to hold the state for the mock config."""

        state_path: Optional[Path] = None
        raise_exception: bool = False
        config_found: bool = True

    mock_config_data = MockConfigContainer()

    def mock_get_config_replacement():
        """Replacement for quam.config.get_quam_config"""
        if not mock_config_data.config_found:
            return None  # Simulate config not found
        if mock_config_data.raise_exception:
            raise RuntimeError("Simulated config loading error")

        # Create a basic object that mimics the necessary structure
        # needed by _get_state_path and its internal calls.
        class MockQuamConfigData:
            # Provide the version attribute expected by validators
            version = EXPECTED_CONFIG_VERSION
            state_path = None  # Initialize state_path

        mock_cfg = MockQuamConfigData()
        mock_cfg.state_path = mock_config_data.state_path  # Set state_path if provided

        return mock_cfg  # Return the object

    # Attempt to mock, handling potential ImportError
    try:
        import quam.config  # noqa: F401

        # Mock the get_quam_config function where it's used
        monkeypatch.setattr(
            "quam.serialisation.json.get_quam_config",
            mock_get_config_replacement,
            raising=False,
        )
    except ImportError:
        # Simulate the ImportError scenario for _get_state_path
        monkeypatch.setattr(
            "quam.serialisation.json.get_quam_config", None, raising=False
        )
        # Also handle the direct import within _get_state_path if it fails
        monkeypatch.setitem(sys.modules, "quam.config", None)

    yield mock_config_data  # Tests can modify mock_config_data.state_path etc.

    # Cleanup
    monkeypatch.undo()
    if "quam.config" in sys.modules and sys.modules["quam.config"] is None:
        del sys.modules["quam.config"]



import pytest

from qm import qua
from quam.components import pulses, SingleChannel


@pytest.fixture
def ch():
    pulse = pulses.GaussianPulse(
        amplitude=1,
        length=10,
        sigma=1,
    )
    channel = SingleChannel(
        id="test_channel",
        operations={"pulse": pulse},
        opx_output=("con1", 1),
    )
    return channel


def test_pulse_no_amplitude_scale(ch):
    with qua.program():
        ch.play("pulse")


def test_pulse_float_amplitude_scale(ch):
    with qua.program():
        ch.play("pulse", amplitude_scale=0.5)


def test_pulse_amplitude_scale_qua_var(ch):
    with qua.program():
        a = qua.declare(qua.fixed)
        ch.play("pulse", amplitude_scale=a)


def test_pulse_amplitude_scale_qua_var_binary_op(ch):
    with qua.program():
        a = qua.declare(qua.fixed)
        ch.play("pulse", amplitude_scale=a * 0.1)


def test_pulse_amplitude_scale_float_list(ch):
    with qua.program():
        ch.play("pulse", amplitude_scale=[0, 0, 0, 1])


def test_pulse_amplitude_scale_float_list_qua_var(ch):
    with qua.program():
        a = qua.declare(qua.fixed)
        ch.play("pulse", amplitude_scale=[0, 0, 0, a])



from abc import ABC
from dataclasses import field
from typing import ClassVar, Dict, List, Optional, Sequence, Literal, Tuple, Union, Any
import warnings
from packaging.version import Version

import qm

from quam.components.hardware import BaseFrequencyConverter, Mixer, LocalOscillator
from quam.components.ports.digital_outputs import (
    DigitalOutputPort,
    OPXPlusDigitalOutputPort,
)
from quam.components.ports.analog_inputs import (
    LFAnalogInputPort,
    LFFEMAnalogInputPort,
    MWFEMAnalogInputPort,
    OPXPlusAnalogInputPort,
)
from quam.components.ports.analog_outputs import (
    LFAnalogOutputPort,
    LFFEMAnalogOutputPort,
    MWFEMAnalogOutputPort,
    OPXPlusAnalogOutputPort,
)
from quam.components.pulses import Pulse, BaseReadoutPulse
from quam.components.ports.digital_outputs import (
    FEMDigitalOutputPort,
)
from quam.core import QuamComponent, quam_dataclass
from quam.core.quam_classes import QuamDict
from quam.utils import string_reference as str_ref
from quam.utils.pulse import add_amplitude_scale_to_pulse_name
from quam.utils.qua_types import (
    ChirpType,
    StreamType,
    ScalarInt,
    ScalarFloat,
    ScalarBool,
    QuaScalarInt,
    QuaVariableInt,
    QuaVariableFloat,
)

from qm.qua import (
    align,
    amp,
    play,
    wait,
    measure,
    declare,
    set_dc_offset,
    fixed,
    demod,
    dual_demod,
    update_frequency,
    frame_rotation,
    frame_rotation_2pi,
    time_tagging,
    reset_if_phase,
)

__all__ = [
    "Channel",
    "DigitalOutputChannel",
    "StickyChannelAddon",
    "TimeTaggingAddon",
    "SingleChannel",
    "InSingleChannel",
    "IQChannel",
    "InIQChannel",
    "InOutSingleChannel",
    "InOutIQChannel",
    "InSingleOutIQChannel",
    "InIQOutSingleChannel",
    "MWChannel",
    "InMWChannel",
    "InOutMWChannel",
]


LF_output_port_types = Union[
    LFFEMAnalogOutputPort,
    OPXPlusAnalogOutputPort,
    Tuple[str, int],
    Tuple[str, int, int],
]

LF_input_port_types = Union[
    LFFEMAnalogInputPort,
    OPXPlusAnalogInputPort,
    Tuple[str, int],
    Tuple[str, int, int],
]


def _create_port_property_deprecation_message(
    channel_class: str,
    property_name: str,
    property_value: Any,
    port_reference: Any,
    port_property_name: Optional[str] = None,
) -> str:
    """Helper function to create deprecation warning message for port properties.

    Args:
        channel_class: Name of the channel class (e.g., "SingleChannel")
        property_name: Name of the deprecated channel property
        property_value: Value of the deprecated property
        port_reference: Port tuple or object being referenced
        port_property_name: Name of the port property (if different from
            channel property). Defaults to inferring from property_name.

    Returns:
        Formatted deprecation warning message string
    """
    # Infer port property name if not provided
    if port_property_name is None:
        if property_name.endswith("_offset") or property_name.endswith("_offset_I") or property_name.endswith("_offset_Q"):
            port_property_name = "offset"
        elif property_name == "filter_fir_taps":
            port_property_name = "feedforward_filter"
        elif property_name == "filter_iir_taps":
            port_property_name = "feedback_filter"
        else:
            port_property_name = property_name  # shareable, inverted, etc.

    # Format port reference for migration example
    if isinstance(port_reference, tuple):
        port_ref_str = repr(port_reference)
    else:
        port_ref_str = f"{port_reference.__class__.__name__}(...)"

    # Build migration message
    message = (
        f"{channel_class}.{property_name} is deprecated and will be removed "
        f"in a future version. Port properties should be configured on "
        f"dedicated Port objects.\n\n"
        f"Migration: Create an explicit port instead:\n"
        f"  from quam.components.ports import OPXPlusAnalogOutputPort\n"
        f"  port = OPXPlusAnalogOutputPort(*{port_ref_str}, "
        f"{port_property_name}={property_value})\n"
        f"  channel = {channel_class}(opx_output=port)\n\n"
        f"See: https://qua-platform.github.io/quam/components/channel-ports/"
    )

    return message


@quam_dataclass
class DigitalOutputChannel(QuamComponent):
    """QUAM component for a digital output channel (signal going out of the OPX)

    Should be added to `Channel.digital_outputs` so that it's also added to the
    respective element in the QUA config.

    Args:
        opx_output (DigitalOutputPort): Channel output port from the OPX perspective,
            E.g. FEMDigitalOutputPort("con1", 1, 2)
        delay (int, optional): Delay in nanoseconds. An intrinsic negative delay of
            136 ns exists by default.
        buffer (int, optional): Digital pulses played to this element will be convolved
            with a digital pulse of value 1 with this length [ns].
        shareable (bool, deprecated): If True, the digital output can be shared with other
            QM instances.
            **Deprecated**: This property has been moved to Port objects. Use
            `OPXPlusDigitalOutputPort(shareable=...)` instead.
            See [Port documentation](channel-ports.md) for details.
        inverted (bool, deprecated): If True, the digital output is inverted.
            **Deprecated**: This property has been moved to Port objects. Use
            `OPXPlusDigitalOutputPort(inverted=...)` instead.
            See [Port documentation](channel-ports.md) for details.
    ."""

    opx_output: Union[Tuple[str, int], Tuple[str, int, int], DigitalOutputPort]
    delay: int = None
    buffer: int = None

    shareable: bool = None
    inverted: bool = None

    def generate_element_config(self) -> Dict[str, int]:
        """Generates the config entry for a digital channel in the QUA config.

        This config entry goes into:
        config.elements.<element_name>.digitalInputs.<opx_output[1]>

        Returns:
            Dict[str, int]: The digital channel config entry.
                Contains "port", and optionally "delay", "buffer" if specified
        """
        if isinstance(self.opx_output, DigitalOutputPort):
            opx_output = self.opx_output.port_tuple
        else:
            opx_output = tuple(self.opx_output)

        digital_cfg: Dict[str, Any] = {"port": opx_output}
        if self.delay is not None:
            digital_cfg["delay"] = self.delay
        if self.buffer is not None:
            digital_cfg["buffer"] = self.buffer
        return digital_cfg

    def apply_to_config(self, config: dict) -> None:
        """Adds this DigitalOutputChannel to the QUA configuration.

        config.controllers.<controller_name>.digital_outputs.<port> will be updated
        with the shareable and inverted settings of this channel if specified.

        See [`QuamComponent.apply_to_config`][quam.core.quam_classes.QuamComponent.apply_to_config]
        for details.
        """
        if isinstance(self.opx_output, DigitalOutputPort):
            if self.shareable is not None:
                msg = _create_port_property_deprecation_message(
                    "DigitalOutputChannel",
                    "shareable",
                    self.shareable,
                    self.opx_output,
                )
                warnings.warn(msg, DeprecationWarning, stacklevel=2)
            if self.inverted is not None:
                msg = _create_port_property_deprecation_message(
                    "DigitalOutputChannel",
                    "inverted",
                    self.inverted,
                    self.opx_output,
                )
                warnings.warn(msg, DeprecationWarning, stacklevel=2)
            return

        # Warn about deprecated port properties when using tuple notation
        if self.shareable is not None:
            msg = _create_port_property_deprecation_message(
                "DigitalOutputChannel",
                "shareable",
                self.shareable,
                self.opx_output,
            )
            warnings.warn(msg, DeprecationWarning, stacklevel=2)

        if self.inverted is not None:
            msg = _create_port_property_deprecation_message(
                "DigitalOutputChannel",
                "inverted",
                self.inverted,
                self.opx_output,
            )
            warnings.warn(msg, DeprecationWarning, stacklevel=2)

        shareable = self.shareable if self.shareable is not None else False
        inverted = self.inverted if self.inverted is not None else False
        if len(self.opx_output) == 2:
            digital_output_port = OPXPlusDigitalOutputPort(
                *self.opx_output, shareable=shareable, inverted=inverted
            )
        else:
            digital_output_port = FEMDigitalOutputPort(
                *self.opx_output, shareable=shareable, inverted=inverted
            )
        digital_output_port.apply_to_config(config)


@quam_dataclass
class StickyChannelAddon(QuamComponent):
    """Addon to make channels sticky.

    Args:
        duration (int): The ramp to zero duration, in ns.
        enabled (bool, optional): If False, the sticky parameters are not applied.
            Default is True.
        analog (bool, optional): If False, the sticky parameters are not applied to
            analog outputs. Default is True.
        digital (bool, optional): If False, the sticky parameters are not applied to
            digital outputs. Default is True.
    """

    duration: int
    enabled: bool = True
    analog: bool = True
    digital: bool = True

    @property
    def channel(self) -> Optional["Channel"]:
        """If the parent is a channel, returns the parent, otherwise returns None."""
        if isinstance(self.parent, Channel):
            return self.parent
        else:
            return

    @property
    def config_settings(self):
        if self.channel is not None:
            return {"after": [self.channel]}

    def apply_to_config(self, config: dict) -> None:
        if self.channel is None:
            return

        if not self.enabled:
            return

        config["elements"][self.channel.name]["sticky"] = {
            "analog": self.analog,
            "digital": self.digital,
            "duration": self.duration,
        }


@quam_dataclass
class TimeTaggingAddon(QuamComponent):
    """Addon to perform time tagging on a channel.

    Args:
        signal_threshold (float, optional): The signal threshold in volts.
            If not specified, the default value is 800 / 4096 ≈ 0.195 V.
        signal_polarity (Literal["above", "below"]): The polarity of the signal
            threshold. Default is "below".
        derivative_threshold (float, optional): The derivative threshold in volts/ns.
            If not specified, the default value is 300 / 4096 ≈ 0.073 V/ns.
        derivative_polarity (Literal["above", "below"]): The polarity of the derivative
            threshold. Default is "below".

    For details see [Time Tagging](https://docs.quantum-machines.co/latest/docs/Guides/features/#time-tagging)
    """

    signal_threshold: float = 800 / 4096
    signal_polarity: Literal["above", "below"] = "below"
    derivative_threshold: float = 300 / 4096
    derivative_polarity: Literal["above", "below"] = "below"
    enabled: bool = True

    @property
    def channel(self) -> Optional["Channel"]:
        """If the parent is a channel, returns the parent, otherwise returns None."""
        if isinstance(self.parent, Channel):
            return self.parent
        else:
            return

    @property
    def config_settings(self):
        if self.channel is not None:
            return {"after": [self.channel]}

    def apply_to_config(self, config: dict) -> None:
        if self.channel is None:
            return

        if not self.enabled:
            return

        if self.signal_threshold is not None and abs(self.signal_threshold) > 1:
            raise ValueError("TimeTaggingAddon.signal_threshold must be a voltage")
        # TODO should we also check derivative threshold? What should the max value be?

        ch_cfg = config["elements"][self.channel.name]
        ch_cfg["timeTaggingParameters"] = {
            "signalThreshold": int(self.signal_threshold * 4096),
            "signalPolarity": self.signal_polarity,
            "derivativeThreshold": int(self.derivative_threshold * 4096),
            "derivativePolarity": self.derivative_polarity,
        }


@quam_dataclass
class Channel(QuamComponent, ABC):
    """Base QUAM component for a channel, can be output, input or both.

    Args:
        operations (Dict[str, Pulse]): A dictionary of pulses to be played on this
            channel. The key is the pulse label (e.g. "X90") and value is a Pulse.
        id (str, int): The id of the channel, used to generate the name.
            Can be a string, or an integer in which case it will add
            `Channel._default_label`.
        sticky (Sticky): Optional sticky parameters for the channel, i.e. defining
            whether successive pulses are applied w.r.t the previous pulse or w.r.t 0 V.
            If not specified, this channel is not sticky.
        digital_outputs (Dict[str, DigitalOutputChannel]): A dictionary of digital
            output channels to be used on this channel. The key is the label of the
            digital output channel (e.g. "DO1") and the value is a DigitalOutputChannel.
        intermediate_frequency (float, optional): The intermediate frequency of the
            channel in Hz. If not specified, the intermediate frequency is zero.
        core (str, optional): The core to use for the channel, useful when sharing a
            core between channels. If not specified, the core is assigned automatically.
        thread (str, optional): The channel core, duplicate of 'core' argument, and
            deprecated from qm.qua >= 1.2.2.
    """

    operations: Dict[str, Pulse] = field(default_factory=dict)

    id: Union[str, int] = None
    _default_label: ClassVar[str] = "ch"  # Used to determine name from id

    digital_outputs: Dict[str, DigitalOutputChannel] = field(default_factory=dict)
    sticky: Optional[StickyChannelAddon] = None
    intermediate_frequency: Optional[float] = None

    thread: Optional[str] = None
    core: Optional[str] = None

    @property
    def name(self) -> str:
        cls_name = self.__class__.__name__

        if self.id is not None:
            if str_ref.is_reference(self.id):
                raise AttributeError(
                    f"{cls_name}.name cannot be determined. "
                    f"Please either set {cls_name}.id to a string or integer, "
                    f"or {cls_name} should be an attribute of another QUAM component."
                )
            if isinstance(self.id, str):
                return self.id
            else:
                return f"{self._default_label}{self.id}"
        if self.parent is None:
            raise AttributeError(
                f"{cls_name}.name cannot be determined. "
                f"Please either set {cls_name}.id to a string or integer, "
                f"or {cls_name} should be an attribute of another QUAM component with "
                "a name."
            )
        if isinstance(self.parent, QuamDict):
            return self.parent.get_attr_name(self)
        if not hasattr(self.parent, "name"):
            raise AttributeError(
                f"{cls_name}.name cannot be determined. "
                f"Please either set {cls_name}.id to a string or integer, "
                f"or {cls_name} should be an attribute of another QUAM component with "
                "a name."
            )
        return f"{self.parent.name}{str_ref.DELIMITER}{self.parent.get_attr_name(self)}"

    @property
    def pulse_mapping(self):
        return {label: pulse.pulse_name for label, pulse in self.operations.items()}

    def play(
        self,
        pulse_name: str,
        amplitude_scale: Optional[Union[ScalarFloat, Sequence[ScalarFloat]]] = None,
        duration: ScalarInt = None,
        condition: ScalarBool = None,
        chirp: ChirpType = None,
        truncate: ScalarInt = None,
        timestamp_stream: StreamType = None,
        continue_chirp: bool = False,
        target: str = "",
        validate: bool = True,
    ):
        """Play a pulse on this channel.

        Args:
            pulse_name (str): The name of the pulse to play. Should be registered in
                `self.operations`.
            amplitude_scale (Optional[Union[ScalarFloat, Sequence[ScalarFloat]]]):
                Amplitude scale of the pulse. Can be either a (qua) float, or a list of
                (qua) floats. If None, the pulse is played without amplitude scaling.
            duration (Scalar[int]): Duration of the pulse in units of the
                clock cycle (4ns). If not provided, the default pulse duration will be
                used. It is possible to dynamically change the duration of both constant
                and arbitrary pulses. Arbitrary pulses can only be stretched, not
                compressed
            chirp (Union[(list[int], str), (int, str)]): Allows to perform
                piecewise linear sweep of the element's intermediate
                frequency in time. Input should be a tuple, with the 1st
                element being a list of rates and the second should be a
                string with the units. The units can be either: 'Hz/nsec',
                'mHz/nsec', 'uHz/nsec', 'pHz/nsec' or 'GHz/sec', 'MHz/sec',
                'KHz/sec', 'Hz/sec', 'mHz/sec'.
            truncate (Scalar[int]): Allows playing
                only part of the pulse, truncating the end. If provided,
                will play only up to the given time in units of the clock
                cycle (4ns).
            condition (A logical expression to evaluate.): Will play analog
                pulse only if the condition's value is true. Any digital
                pulses associated with the operation will always play.
            timestamp_stream (Union[str, _ResultSource]): (Supported from
                QOP 2.2) Adding a `timestamp_stream` argument will save the
                time at which the operation occurred to a stream. If the
                `timestamp_stream` is a string ``label``, then the timestamp
                handle can be retrieved with
                `qm._results.JobResults.get` with the same ``label``.
            validate (bool): If True (default), validate that the pulse is registered
                in Channel.operations

        Note:
            The `element` argument from `qm.qua.play()`is not needed, as it is
            automatically set to `self.name`.

        """
        if validate and pulse_name not in self.operations:
            raise KeyError(
                f"Operation '{pulse_name}' not found in channel '{self.name}'"
            )

        pulse_name_with_amp_scale = add_amplitude_scale_to_pulse_name(
            pulse_name, amplitude_scale
        )

        # At the moment, self.name is not defined for Channel because it could
        # be a property or dataclass field in a subclass.
        # # TODO Find elegant solution for Channel.name.
        play(
            pulse=pulse_name_with_amp_scale,
            element=self.name,
            duration=duration,
            condition=condition,
            chirp=chirp,
            truncate=truncate,
            timestamp_stream=timestamp_stream,
            continue_chirp=continue_chirp,
            target=target,
        )

    def wait(self, duration: ScalarInt, *other_elements: Union[str, "Channel"]):
        """Wait for the given duration on all provided elements without outputting anything.

        Duration is in units of the clock cycle (4ns)

        Args:
            duration (Scalar[int]): time to wait in units of the clock cycle
                (4ns). Range: [4, $2^{31}-1$] in steps of 1.
            *other_elements (Union[str,sequence of str]): elements to wait on,
                in addition to this channel

        Warning:
            In case the value of this is outside the range above, unexpected results may occur.

        Note:
            The current channel element is always included in the wait operation.

        Note:
            The purpose of the `wait` operation is to add latency. In most cases, the
            latency added will be exactly the same as that specified by the QUA variable or
            the literal used. However, in some cases an additional computational latency may
            be added. If the actual wait time has significance, such as in characterization
            experiments, the actual wait time should always be verified with a simulator.
        """
        other_elements_str = [
            element if isinstance(element, str) else str(element)
            for element in other_elements
        ]
        wait(duration, self.name, *other_elements_str)

    def align(self, *other_elements):
        if not other_elements:
            align()
        else:
            other_elements_str = [
                element if isinstance(element, str) else str(element)
                for element in other_elements
            ]
            align(self.name, *other_elements_str)

    def update_frequency(
        self,
        new_frequency: ScalarInt,
        units: str = "Hz",
        keep_phase: bool = False,
    ):
        """Dynamically update the frequency of the associated oscillator.

        This changes the frequency from the value defined in the channel.

        The behavior of the phase (continuous vs. coherent) is controlled by the
        ``keep_phase`` parameter and is discussed in the documentation.

        Args:
            new_frequency (Scalar[int]): The new frequency value to set
                in units set by ``units`` parameter. In steps of 1.
            units (str): units of new frequency. Useful when sub-Hz
                precision is required. Allowed units are "Hz", "mHz", "uHz",
                "nHz", "pHz"
            keep_phase (bool): Determine whether phase will be continuous
                through the change (if ``True``) or it will be coherent,
                only the frequency will change (if ``False``).

        Example:
            ```python
            with program() as prog:
                update_frequency("q1", 4e6) # will set the frequency to 4 MHz

                ### Example for sub-Hz resolution
                # will set the frequency to 100 Hz (due to casting to int)
                update_frequency("q1", 100.7)

                # will set the frequency to 100.7 Hz
                update_frequency("q1", 100700, units='mHz')
            ```
        """
        update_frequency(self.name, new_frequency, units, keep_phase)

    def reset_if_phase(self):
        r"""
        Resets the intermediate frequency phase of the oscillator, setting the phase of
        the next pulse to absolute zero.
        This sets the phase of the currently playing intermediate frequency
        to the value it had at the beginning of the program (t=0).

        Note:
        - The phase will only be set to zero when the next play or align command is
          executed on the element.
        - Reset phase will only reset the phase of the intermediate frequency
          (:math:`\\omega_{IF}`) currently in use.
        """
        reset_if_phase(self.name)

    def frame_rotation(self, angle: ScalarFloat):
        r"""Shift the phase of the channel element's oscillator by the given angle.

        This is typically used for virtual z-rotations.

        Note:
            The fixed point format of QUA variables of type fixed is 4.28, meaning the
            phase must be between $-8$ and $8-2^{28}$. Otherwise the phase value will be
            invalid. It is therefore better to use `frame_rotation_2pi()` which avoids
            this issue.

        Note:
            The phase is accumulated with a resolution of 16 bit.
            Therefore, *N* changes to the phase can result in a phase (and amplitude)
            inaccuracy of about :math:`N \cdot 2^{-16}`. To null out this accumulated
            error, it is recommended to use `reset_frame(el)` from time to time.

        Args:
            angle (Scalar[float]): The angle to add to the current
                phase (in radians)
            *elements (str): a single element whose oscillator's phase will
                be shifted. multiple elements can be given, in which case
                all of their oscillators' phases will be shifted

        """
        frame_rotation(angle, self.name)

    def frame_rotation_2pi(self, angle: ScalarFloat):
        r"""Shift the phase of the oscillator associated with an element by the given
        angle in units of 2pi radians.

        This is typically used for virtual z-rotations.

        Note:
            Unlike the case of frame_rotation(), this method performs the 2-pi radian
            wrap around of the angle automatically.

        Note:
            The phase is accumulated with a resolution of 16 bit.
            Therefore, *N* changes to the phase can result in a phase inaccuracy of
            about :math:`N \cdot 2^{-16}`. To null out this accumulated error, it is
            recommended to use `reset_frame(el)` from time to time.

        Args:
            angle (Scalar[float]): The angle to add to the current
                phase (in $2\pi$ radians)
        """
        frame_rotation_2pi(angle, self.name)

    def _config_add_digital_outputs(self, config: Dict[str, dict]) -> None:
        """Adds the digital outputs to the QUA config.

        config.elements.<element_name>.digitalInputs will be updated with the digital
        outputs of this channel.

        Note that the digital outputs are added separately to the controller config in
        `DigitalOutputChannel.apply_to_config`.

        Args:
            config (dict): The QUA config that's in the process of being generated.
        """
        if not self.digital_outputs:
            return

        element_config = config["elements"][self.name]
        element_config.setdefault("digitalInputs", {})

        for name, digital_output in self.digital_outputs.items():
            digital_cfg = digital_output.generate_element_config()
            element_config["digitalInputs"][name] = digital_cfg

    def apply_to_config(self, config: Dict[str, dict]) -> None:
        """Adds this Channel to the QUA configuration.

        config.elements.<element_name> will be created, and the operations are added.

        Args:
            config (dict): The QUA config that's in the process of being generated.

        Raises:
            ValueError: If the channel already exists in the config.
        """
        if self.name in config["elements"]:
            raise ValueError(
                f"Cannot add channel '{self.name}' to the config because it already "
                f"exists. Existing entry: {config['elements'][self.name]}"
            )
        config["elements"][self.name] = {"operations": self.pulse_mapping}
        element_config = config["elements"][self.name]

        if self.intermediate_frequency is not None:
            element_config["intermediate_frequency"] = self.intermediate_frequency

        try:
            qua_below_1_2_2 = Version(qm.__version__) <= Version("1.2.1")
        except ImportError:
            warnings.warn(
                "Unable to to determine qm package version, assuming < 1.2.2. "
            )
            qua_below_1_2_2 = True

        if self.core is not None and self.thread is not None:
            warnings.warn(
                "The 'thread' and 'core' arguments are mutually exclusive. "
                "Using 'core' instead."
            )
            core = self.core
        elif self.thread is not None:
            if not qua_below_1_2_2:
                warnings.warn(
                    "The 'thread' element argument is deprecated from qm.qua >= 1.2.2. "
                    "Use 'core' instead."
                )
            core = self.thread
        else:
            core = self.core

        if core is not None:
            if qua_below_1_2_2:
                element_config["thread"] = core
            else:
                element_config["core"] = core

        self._config_add_digital_outputs(config)


@quam_dataclass
class SingleChannel(Channel):
    """QUAM component for a single (not IQ) output channel.

    Args:
        operations (Dict[str, Pulse]): A dictionary of pulses to be played on this
            channel. The key is the pulse label (e.g. "X90") and value is a Pulse.
        id (str, int): The id of the channel, used to generate the name.
            Can be a string, or an integer in which case it will add
            `Channel._default_label`.
        opx_output (LF_output_port_types): Channel output port from the OPX perspective,
            E.g. LFFEMAnalogOutputPort("con1", 1, 2)
        filter_fir_taps (List[float], deprecated): FIR filter taps for the output port.
            **Deprecated**: This property has been moved to Port objects. Use
            `OPXPlusAnalogOutputPort(feedforward_filter=...)` instead.
            See [Port documentation](channel-ports.md) for details.
        filter_iir_taps (List[float], deprecated): IIR filter taps for the output port.
            **Deprecated**: This property has been moved to Port objects. Use
            `OPXPlusAnalogOutputPort(feedback_filter=...)` instead.
            See [Port documentation](channel-ports.md) for details.
        opx_output_offset (float, deprecated): DC offset for the output port.
            **Deprecated**: This property has been moved to Port objects. Use
            `OPXPlusAnalogOutputPort(offset=...)` instead.
            See [Port documentation](channel-ports.md) for details.
        intermediate_frequency (float): Intermediate frequency of OPX output, default
            is None.
    """

    opx_output: LF_output_port_types
    filter_fir_taps: List[float] = None
    filter_iir_taps: List[float] = None

    opx_output_offset: float = None

    def set_dc_offset(self, offset: ScalarFloat):
        """Set the DC offset of an element's input to the given value.
        This value will remain the DC offset until changed or until the Quantum Machine
        is closed.

        Args:
            offset (Scalar[float]): The DC offset to set the input to.
                This is limited by the OPX output voltage range.
        """
        set_dc_offset(element=self.name, element_input="single", offset=offset)

    def apply_to_config(self, config: dict):
        """Adds this SingleChannel to the QUA configuration.

        See [`QuamComponent.apply_to_config`][quam.core.quam_classes.QuamComponent.apply_to_config]
        for details.
        """
        # Add pulses & waveforms
        super().apply_to_config(config)

        if str_ref.is_reference(self.name):
            raise AttributeError(
                f"Channel {self.get_reference()} cannot be added to the config because"
                " it doesn't have a name. Either set channel.id to a string or"
                " integer, or channel should be an attribute of another QUAM component"
                " with a name."
            )

        element_config = config["elements"][self.name]

        # Check for deprecated port properties on channel
        if self.opx_output_offset is not None:
            msg = _create_port_property_deprecation_message(
                "SingleChannel",
                "opx_output_offset",
                self.opx_output_offset,
                self.opx_output,
            )
            warnings.warn(msg, DeprecationWarning, stacklevel=2)

        filter_fir_taps = self.filter_fir_taps
        if filter_fir_taps is not None:
            msg = _create_port_property_deprecation_message(
                "SingleChannel",
                "filter_fir_taps",
                filter_fir_taps,
                self.opx_output,
            )
            warnings.warn(msg, DeprecationWarning, stacklevel=2)
            filter_fir_taps = list(filter_fir_taps)
        filter_iir_taps = self.filter_iir_taps
        if filter_iir_taps is not None:
            msg = _create_port_property_deprecation_message(
                "SingleChannel",
                "filter_iir_taps",
                filter_iir_taps,
                self.opx_output,
            )
            warnings.warn(msg, DeprecationWarning, stacklevel=2)
            filter_iir_taps = list(filter_iir_taps)

        if isinstance(self.opx_output, LFAnalogOutputPort):
            opx_port = self.opx_output
        elif len(self.opx_output) == 2:
            opx_port = OPXPlusAnalogOutputPort(
                *self.opx_output,
                offset=self.opx_output_offset,
                feedforward_filter=filter_fir_taps,
                feedback_filter=filter_iir_taps,
            )
            opx_port.apply_to_config(config)
        else:
            opx_port = LFFEMAnalogOutputPort(
                *self.opx_output,
                offset=self.opx_output_offset,
                feedforward_filter=filter_fir_taps,
                feedback_filter=filter_iir_taps,
            )
            opx_port.apply_to_config(config)

        element_config["singleInput"] = {"port": opx_port.port_tuple}


@quam_dataclass
class InSingleChannel(Channel):
    """QUAM component for a single (not IQ) input channel.

    Args:
        operations (Dict[str, Pulse]): A dictionary of pulses to be played on this
            channel. The key is the pulse label (e.g. "X90") and value is a Pulse.
        id (str, int): The id of the channel, used to generate the name.
            Can be a string, or an integer in which case it will add
            `Channel._default_label`.
        opx_input (LF_input_port_types): Channel input port from OPX perspective,
            E.g. LFFEMAnalogInputPort("con1", 1, 2)
        opx_input_offset (float, deprecated): DC offset for the input port.
            **Deprecated**: This property has been moved to Port objects. Use
            `OPXPlusAnalogInputPort(offset=...)` instead.
            See [Port documentation](channel-ports.md) for details.
        intermediate_frequency (float): Intermediate frequency of OPX input,
            default is None.
        time_of_flight (int): Round-trip signal duration in nanoseconds.
        smearing (int): Additional window of ADC integration in nanoseconds.
            Used to account for signal smearing.
    """

    opx_input: LF_input_port_types
    opx_input_offset: float = None

    time_of_flight: int = 140
    smearing: int = 0

    time_tagging: Optional[TimeTaggingAddon] = None

    def apply_to_config(self, config: dict):
        """Adds this InSingleChannel to the QUA configuration.

        See [`QuamComponent.apply_to_config`][quam.core.quam_classes.QuamComponent.apply_to_config]
        for details.
        """
        # Add output to config
        super().apply_to_config(config)

        # Note outputs instead of inputs because it's w.r.t. the QPU
        element_config = config["elements"][self.name]
        element_config["smearing"] = self.smearing
        element_config["time_of_flight"] = self.time_of_flight

        # Check for deprecated port properties on channel
        if self.opx_input_offset is not None:
            msg = _create_port_property_deprecation_message(
                "InSingleChannel",
                "opx_input_offset",
                self.opx_input_offset,
                self.opx_input,
            )
            warnings.warn(msg, DeprecationWarning, stacklevel=2)

        if isinstance(self.opx_input, LFAnalogInputPort):
            opx_port = self.opx_input
        elif len(self.opx_input) == 2:
            opx_port = OPXPlusAnalogInputPort(
                *self.opx_input, offset=self.opx_input_offset
            )
            opx_port.apply_to_config(config)
        else:
            opx_port = LFFEMAnalogInputPort(
                *self.opx_input, offset=self.opx_input_offset
            )
            opx_port.apply_to_config(config)

        element_config["outputs"] = {"out1": opx_port.port_tuple}

    def measure(
        self,
        pulse_name: str,
        amplitude_scale: Optional[Union[ScalarFloat, Sequence[ScalarFloat]]] = None,
        qua_vars: Tuple[QuaVariableFloat, ...] = None,
        stream=None,
    ) -> Tuple[QuaVariableFloat, QuaVariableFloat]:
        """Perform a full demodulation measurement on this channel.

        Args:
            pulse_name (str): The name of the pulse to play. Should be registered in
                `self.operations`.
            amplitude_scale (Optional[Union[ScalarFloat, Sequence[ScalarFloat]]]):
                Amplitude scale of the pulse. Can be either a (qua) float, or a list of
                (qua) floats. If None, the pulse is played without amplitude scaling.
            qua_vars (Tuple[QuaVariable[float], ...], optional): Two QUA
                variables to store the I, Q measurement results.
                If not provided, new variables will be declared and returned.
            stream (Optional[StreamType]): The stream to save the measurement result to.
                If not provided, the raw ADC signal will not be streamed.

        Returns:
            I, Q: The QUA variables used to store the measurement results.
                If provided as input, the same variables will be returned.
                If not provided, new variables will be declared and returned.

        Raises:
            ValueError: If `qua_vars` is provided and is not a tuple of two QUA
                variables.
        """

        pulse: BaseReadoutPulse = self.operations[pulse_name]

        if qua_vars is not None:
            if not isinstance(qua_vars, Sequence) or len(qua_vars) != 2:
                raise ValueError(
                    f"InOutSingleChannel.measure received kwarg 'qua_vars' "
                    f"which is not a tuple of two QUA variables. Received {qua_vars=}"
                )
        else:
            qua_vars = [declare(fixed) for _ in range(2)]

        pulse_name_with_amp_scale = add_amplitude_scale_to_pulse_name(
            pulse_name, amplitude_scale
        )

        integration_weight_labels = list(pulse.integration_weights_mapping)
        measure(
            pulse_name_with_amp_scale,
            self.name,
            demod.full(integration_weight_labels[0], qua_vars[0], "out1"),
            demod.full(integration_weight_labels[1], qua_vars[1], "out1"),
            adc_stream=stream,
        )
        return tuple(qua_vars)

    def measure_accumulated(
        self,
        pulse_name: str,
        amplitude_scale: Optional[Union[ScalarFloat, Sequence[ScalarFloat]]] = None,
        num_segments: int = None,
        segment_length: int = None,
        qua_vars: Tuple[QuaVariableFloat, ...] = None,
        stream=None,
    ) -> Tuple[QuaVariableFloat, QuaVariableFloat]:
        """Perform an accumulated demodulation measurement on this channel.

        Args:
            pulse_name (str): The name of the pulse to play. Should be registered in
                `self.operations`.
            amplitude_scale (Optional[Union[ScalarFloat, Sequence[ScalarFloat]]]):
                Amplitude scale of the pulse. Can be either a (qua) float, or a list of
                (qua) floats. If None, the pulse is played without amplitude scaling.
            num_segments (int): The number of segments to accumulate.
                Should either specify this or `segment_length`.
            segment_length (int): The length of the segment to accumulate.
                Should either specify this or `num_segments`.
            qua_vars (Tuple[QuaVariableFloat, ...], optional): Two QUA
                variables to store the I, Q measurement results.
                If not provided, new variables will be declared and returned.
            stream (Optional[StreamType]): The stream to save the measurement result to.
                If not provided, the raw ADC signal will not be streamed.

        Returns:
            I, Q: The QUA variables used to store the measurement results.
                If provided as input, the same variables will be returned.
                If not provided, new variables will be declared and returned.

        Raises:
            ValueError: If both `num_segments` and `segment_length` are provided, or if
                neither are provided.
            ValueError: If `qua_vars` is provided and is not a tuple of two QUA
                variables.
        """
        pulse: BaseReadoutPulse = self.operations[pulse_name]

        if num_segments is None and segment_length is None:
            raise ValueError(
                "InOutSingleChannel.measure_accumulated requires either 'segment_length' "
                "or 'num_segments' to be provided."
            )
        elif num_segments is not None and segment_length is not None:
            raise ValueError(
                "InOutSingleChannel.measure_accumulated received both 'segment_length' "
                "and 'num_segments'. Please provide only one."
            )
        elif num_segments is None:
            num_segments = int(pulse.length / (4 * segment_length))  # Number of slices
        elif segment_length is None:
            segment_length = int(pulse.length / (4 * num_segments))

        if qua_vars is not None:
            if not isinstance(qua_vars, Sequence) or len(qua_vars) != 2:
                raise ValueError(
                    f"InOutSingleChannel.measure_accumulated received kwarg 'qua_vars' "
                    f"which is not a tuple of two QUA variables. Received {qua_vars=}"
                )
        else:
            qua_vars = [declare(fixed, size=num_segments) for _ in range(2)]

        pulse_name_with_amp_scale = add_amplitude_scale_to_pulse_name(
            pulse_name, amplitude_scale
        )

        integration_weight_labels = list(pulse.integration_weights_mapping)
        measure(
            pulse_name_with_amp_scale,
            self.name,
            demod.accumulated(
                integration_weight_labels[0], qua_vars[0], segment_length, "out1"
            ),
            demod.accumulated(
                integration_weight_labels[1], qua_vars[1], segment_length, "out1"
            ),
            adc_stream=stream,
        )
        return tuple(qua_vars)

    def measure_sliced(
        self,
        pulse_name: str,
        amplitude_scale: Optional[Union[ScalarFloat, Sequence[ScalarFloat]]] = None,
        num_segments: int = None,
        segment_length: int = None,
        qua_vars: Tuple[QuaVariableFloat, ...] = None,
        stream=None,
    ) -> Tuple[QuaVariableFloat, QuaVariableFloat]:
        """Perform an accumulated demodulation measurement on this channel.

        Args:
            pulse_name (str): The name of the pulse to play. Should be registered in
                `self.operations`.
            amplitude_scale (Optional[Union[ScalarFloat, Sequence[ScalarFloat]]]):
                Amplitude scale of the pulse. Can be either a (qua) float, or a list of
                (qua) floats. If None, the pulse is played without amplitude scaling.
            num_segments (int): The number of segments to accumulate.
                Should either specify this or `segment_length`.
            segment_length (int): The length of the segment to accumulate.
                Should either specify this or `num_segments`.
            qua_vars (Tuple[QuaVariableFloat, ...], optional): Two QUA
                variables to store the I, Q measurement results.
                If not provided, new variables will be declared and returned.
            stream (Optional[StreamType]): The stream to save the measurement result to.
                If not provided, the raw ADC signal will not be streamed.

        Returns:
            I, Q: The QUA variables used to store the measurement results.
                If provided as input, the same variables will be returned.
                If not provided, new variables will be declared and returned.

        Raises:
            ValueError: If both `num_segments` and `segment_length` are provided, or if
                neither are provided.
            ValueError: If `qua_vars` is provided and is not a tuple of two QUA
                variables.
        """
        pulse: BaseReadoutPulse = self.operations[pulse_name]

        if num_segments is None and segment_length is None:
            raise ValueError(
                "InOutSingleChannel.measure_sliced requires either 'segment_length' "
                "or 'num_segments' to be provided."
            )
        elif num_segments is not None and segment_length is not None:
            raise ValueError(
                "InOutSingleChannel.measure_sliced received both 'segment_length' "
                "and 'num_segments'. Please provide only one."
            )
        elif num_segments is None:
            num_segments = int(pulse.length / (4 * segment_length))  # Number of slices
        elif segment_length is None:
            segment_length = int(pulse.length / (4 * num_segments))

        if qua_vars is not None:
            if not isinstance(qua_vars, Sequence) or len(qua_vars) != 2:
                raise ValueError(
                    f"InOutSingleChannel.measure_sliced received kwarg 'qua_vars' "
                    f"which is not a tuple of two QUA variables. Received {qua_vars=}"
                )
        else:
            qua_vars = [declare(fixed, size=num_segments) for _ in range(2)]

        pulse_name_with_amp_scale = add_amplitude_scale_to_pulse_name(
            pulse_name, amplitude_scale
        )

        integration_weight_labels = list(pulse.integration_weights_mapping)
        measure(
            pulse_name_with_amp_scale,
            self.name,
            demod.sliced(
                integration_weight_labels[0], qua_vars[0], segment_length, "out1"
            ),
            demod.sliced(
                integration_weight_labels[1], qua_vars[1], segment_length, "out1"
            ),
            adc_stream=stream,
        )
        return tuple(qua_vars)

    def measure_time_tagging(
        self,
        pulse_name: str,
        size: int,
        max_time: int,
        qua_vars: Optional[Tuple[QuaVariableInt, QuaScalarInt]] = None,
        stream: Optional[StreamType] = None,
        mode: Literal["analog", "high_res", "digital"] = "analog",
    ) -> Tuple[QuaVariableInt, QuaScalarInt]:
        """Perform a time tagging measurement on this channel.

        For details see https://docs.quantum-machines.co/latest/docs/Guides/features/#time-tagging

        Args:
            pulse_name (str): The name of the pulse to play. Should be registered in
                `self.operations`.
            size (int): The size of the QUA array to store the times of the detected
                pulses. Ignored if `qua_vars` is provided.
            max_time (int): The maximum time to search for pulses.
            qua_vars (Tuple[QuaVariableInt, QuaScalarInt], optional): QUA variables
                to store the times and counts of the detected pulses. If not provided,
                new variables will be declared and returned.
            stream (Optional[StreamType]): The stream to save the measurement result to.
                If not provided, the raw ADC signal will not be streamed.
            mode (Literal["analog", "high_res", "digital"]): The time tagging mode.

        Returns:
            times (QuaVariable[Any]): The QUA variable to store the times of the detected
                pulses.
            counts (QuaScalar[int]): The number of detected pulses.

        Example:
            ```python
            times, counts = channel.measure_time_tagging("readout", size=1000, max_time=1000)
            ```
        """
        if mode == "analog":
            time_tagging_func = time_tagging.analog
        elif mode == "high_res":
            time_tagging_func = time_tagging.high_res
        elif mode == "digital":
            time_tagging_func = time_tagging.digital
        else:
            raise ValueError(f"Invalid time tagging mode: {mode}")

        if qua_vars is None:
            times = declare(int, size=size)
            counts = declare(int)
        else:
            times, counts = qua_vars

        measure(
            pulse_name,
            self.name,
            time_tagging_func(target=times, max_time=max_time, targetLen=counts),
            adc_stream=stream,
        )
        return times, counts


@quam_dataclass
class _OutComplexChannel(Channel, ABC):
    """Base class for IQ and MW output channels."""

    LO_frequency: float
    RF_frequency: float

    @property
    def inferred_RF_frequency(self) -> float:
        """Inferred RF frequency by adding LO and IF

        Can be used by having reference `RF_frequency = "#./inferred_RF_frequency"`
        Returns:
            self.LO_frequency + self.intermediate_frequency
        """
        name = getattr(self, "name", self.__class__.__name__)
        if not isinstance(self.LO_frequency, (float, int)):
            raise AttributeError(
                f"Error inferring RF frequency for channel {name}: "
                f"LO_frequency is not a number: {self.LO_frequency}"
            )
        if not isinstance(self.intermediate_frequency, (float, int)):
            raise AttributeError(
                f"Error inferring RF frequency for channel {name}: "
                f"intermediate_frequency is not a number: {self.intermediate_frequency}"
            )
        return self.LO_frequency + self.intermediate_frequency

    @property
    def inferred_intermediate_frequency(self) -> float:
        """Inferred intermediate frequency by subtracting LO from RF

        Can be used by having reference
        `intermediate_frequency = "#./inferred_intermediate_frequency"`

        Returns:
            self.RF_frequency - self.LO_frequency
        """
        name = getattr(self, "name", self.__class__.__name__)
        if not isinstance(self.LO_frequency, (float, int)):
            raise AttributeError(
                f"Error inferring intermediate frequency for channel {name}: "
                f"LO_frequency is not a number: {self.LO_frequency}"
            )
        if not isinstance(self.RF_frequency, (float, int)):
            raise AttributeError(
                f"Error inferring intermediate frequency for channel {name}: "
                f"RF_frequency is not a number: {self.RF_frequency}"
            )
        return self.RF_frequency - self.LO_frequency

    @property
    def inferred_LO_frequency(self) -> float:
        """Inferred LO frequency by subtracting IF from RF

        Can be used by having reference `LO_frequency = "#./inferred_LO_frequency"`

        Returns:
            self.RF_frequency - self.intermediate_frequency
        """
        name = getattr(self, "name", self.__class__.__name__)
        if not isinstance(self.RF_frequency, (float, int)):
            raise AttributeError(
                f"Error inferring LO frequency for channel {name}: "
                f"RF_frequency is not a number: {self.RF_frequency}"
            )
        if not isinstance(self.intermediate_frequency, (float, int)):
            raise AttributeError(
                f"Error inferring LO frequency for channel {name}: "
                f"intermediate_frequency is not a number: {self.intermediate_frequency}"
            )
        return self.RF_frequency - self.intermediate_frequency


@quam_dataclass
class IQChannel(_OutComplexChannel):
    """QUAM component for an IQ output channel.

    Args:
        operations (Dict[str, Pulse]): A dictionary of pulses to be played on this
            channel. The key is the pulse label (e.g. "X90") and value is a Pulse.
        id (str, int): The id of the channel, used to generate the name.
            Can be a string, or an integer in which case it will add
            `Channel._default_label`.
        opx_output_I (LF_output_port_types): Channel I output port from the OPX
            perspective, E.g. LFFEMAnalogOutputPort("con1", 1, 1)
        opx_output_Q (LF_output_port_types): Channel Q output port from the OPX
            perspective, E.g. LFFEMAnalogOutputPort("con1", 1, 2)
        opx_output_offset_I (float, deprecated): The offset of the I channel.
            **Deprecated**: This property has been moved to Port objects. Use
            `OPXPlusAnalogOutputPort(offset=...)` on the I port instead.
            See [Port documentation](channel-ports.md) for details.
        opx_output_offset_Q (float, deprecated): The offset of the Q channel.
            **Deprecated**: This property has been moved to Port objects. Use
            `OPXPlusAnalogOutputPort(offset=...)` on the Q port instead.
            See [Port documentation](channel-ports.md) for details.
        intermediate_frequency (float): Intermediate frequency of the mixer.
            Default is 0.0
        LO_frequency (float): Local oscillator frequency. Default is the LO frequency
            of the frequency converter up component.
        RF_frequency (float): RF frequency of the mixer. By default, the RF frequency
            is inferred by adding the LO frequency and the intermediate frequency.
        frequency_converter_up (FrequencyConverter): Frequency converter QUAM component
            for the IQ output.
    """

    opx_output_I: LF_output_port_types
    opx_output_Q: LF_output_port_types

    opx_output_offset_I: float = None
    opx_output_offset_Q: float = None

    frequency_converter_up: BaseFrequencyConverter

    LO_frequency: float = "#./frequency_converter_up/LO_frequency"
    RF_frequency: float = "#./inferred_RF_frequency"

    _default_label: ClassVar[str] = "IQ"

    @property
    def local_oscillator(self) -> Optional[LocalOscillator]:
        return getattr(self.frequency_converter_up, "local_oscillator", None)

    @property
    def mixer(self) -> Optional[Mixer]:
        return getattr(self.frequency_converter_up, "mixer", None)

    @property
    def rf_frequency(self):
        warnings.warn(
            "rf_frequency is deprecated, use RF_frequency instead", DeprecationWarning
        )
        return self.frequency_converter_up.LO_frequency + self.intermediate_frequency

    def set_dc_offset(self, offset: ScalarFloat, element_input: Literal["I", "Q"]):
        """Set the DC offset of an element's input to the given value.
        This value will remain the DC offset until changed or until the Quantum Machine
        is closed.

        Args:
            offset (Scalar[float]): The DC offset to set the input to.
                This is limited by the OPX output voltage range.
            element_input (Literal["I", "Q"]): The element input to set the offset for.

        Raises:
            ValueError: If element_input is not "I" or "Q"
        """
        if element_input not in ["I", "Q"]:
            raise ValueError(
                f"element_input should be either 'I' or 'Q', got {element_input}"
            )
        set_dc_offset(element=self.name, element_input=element_input, offset=offset)

    def apply_to_config(self, config: dict):
        """Adds this IQChannel to the QUA configuration.

        See [`QuamComponent.apply_to_config`][quam.core.quam_classes.QuamComponent.apply_to_config]
        for details.
        """
        # Add pulses & waveforms
        super().apply_to_config(config)

        if str_ref.is_reference(self.name):
            raise AttributeError(
                f"Channel {self.get_reference()} cannot be added to the config because"
                " it doesn't have a name. Either set channel.id to a string or"
                " integer, or channel should be an attribute of another QUAM component"
                " with a name."
            )

        element_config = config["elements"][self.name]

        # Check for deprecated port properties on channel
        if self.opx_output_offset_I is not None:
            msg = _create_port_property_deprecation_message(
                "IQChannel",
                "opx_output_offset_I",
                self.opx_output_offset_I,
                self.opx_output_I,
            )
            warnings.warn(msg, DeprecationWarning, stacklevel=2)

        if self.opx_output_offset_Q is not None:
            msg = _create_port_property_deprecation_message(
                "IQChannel",
                "opx_output_offset_Q",
                self.opx_output_offset_Q,
                self.opx_output_Q,
            )
            warnings.warn(msg, DeprecationWarning, stacklevel=2)

        from quam.components.octave import OctaveUpConverter

        if isinstance(self.frequency_converter_up, OctaveUpConverter):
            octave = self.frequency_converter_up.octave
            if octave is None:
                raise ValueError(
                    f"Error generating config: channel {self.name} has an "
                    f"OctaveUpConverter (id={self.frequency_converter_up.id}) without "
                    "an attached Octave"
                )
            element_config["RF_inputs"] = {
                "port": (octave.name, self.frequency_converter_up.id)
            }
        elif str_ref.is_reference(self.frequency_converter_up):
            raise ValueError(
                f"Error generating config: channel {self.name} could not determine "
                f'"frequency_converter_up", it seems to point to a non-existent '
                f"reference: {self.frequency_converter_up}"
            )
        else:
            element_config["mixInputs"] = {}  # To be filled in next section
            if self.mixer is not None:
                element_config["mixInputs"]["mixer"] = self.mixer.name
            if self.local_oscillator is not None:
                element_config["mixInputs"]["lo_frequency"] = (
                    self.local_oscillator.frequency
                )

        opx_outputs = [self.opx_output_I, self.opx_output_Q]
        offsets = [self.opx_output_offset_I, self.opx_output_offset_Q]
        for I_or_Q, opx_output, offset in zip("IQ", opx_outputs, offsets):
            if isinstance(opx_output, LFAnalogOutputPort):
                opx_port = opx_output
            elif len(opx_output) == 2:
                opx_port = OPXPlusAnalogOutputPort(*opx_output, offset=offset)
                opx_port.apply_to_config(config)
            else:
                opx_port = LFFEMAnalogOutputPort(*opx_output, offset=offset)
                opx_port.apply_to_config(config)

            if "mixInputs" in element_config:
                element_config["mixInputs"][I_or_Q] = opx_port.port_tuple


@quam_dataclass
class _InComplexChannel(Channel, ABC):
    """A specialized channel class for performing complex demodulation measurements.

    This class extends the basic `Channel` class and provides functionality
    for performing full dual demodulation measurements on a channel. It allows
    for the measurement of both in-phase (I) and quadrature (Q) components of
    a signal, which are essential for characterizing quantum signals.

    This class is used for both input IQ channels on low-frequency analog inputs and for
    MW inputs in the MW FEM.
    """

    def measure(
        self,
        pulse_name: str,
        amplitude_scale: Optional[Union[ScalarFloat, Sequence[ScalarFloat]]] = None,
        qua_vars: Tuple[QuaVariableFloat, QuaVariableFloat] = None,
        stream=None,
    ) -> Tuple[QuaVariableFloat, QuaVariableFloat]:
        """Perform a full dual demodulation measurement on this channel.

        Args:
            pulse_name (str): The name of the pulse to play. Should be registered in
                `self.operations`.
            amplitude_scale (Optional[Union[ScalarFloat, Sequence[ScalarFloat]]]):
                Amplitude scale of the pulse. Can be either a (qua) float, or a list of
                (qua) floats. If None, the pulse is played without amplitude scaling.
            qua_vars (Tuple[QuaVariable[float], QuaVariable[float]], optional): Two QUA
                variables to store the I and Q measurement results. If not provided,
                new variables will be declared and returned.
            stream (Optional[StreamType]): The stream to save the measurement result to.
                If not provided, the raw ADC signal will not be streamed.

        Returns:
            I, Q: The QUA variables used to store the measurement results.
                If provided as input, the same variables will be returned.
                If not provided, new variables will be declared and returned.
        """
        pulse: BaseReadoutPulse = self.operations[pulse_name]

        if qua_vars is not None:
            if not isinstance(qua_vars, Sequence) or len(qua_vars) != 2:
                raise ValueError(
                    f"InOutIQChannel.measure received kwarg 'qua_vars' which is not a "
                    f"tuple of two QUA variables. Received {qua_vars=}"
                )
        else:
            qua_vars = [declare(fixed) for _ in range(2)]

        pulse_name_with_amp_scale = add_amplitude_scale_to_pulse_name(
            pulse_name, amplitude_scale
        )

        integration_weight_labels = list(pulse.integration_weights_mapping)
        measure(
            pulse_name_with_amp_scale,
            self.name,
            dual_demod.full(
                iw1=integration_weight_labels[0],
                element_output1="out1",
                iw2=integration_weight_labels[1],
                element_output2="out2",
                target=qua_vars[0],
            ),
            dual_demod.full(
                iw1=integration_weight_labels[2],
                element_output1="out1",
                iw2=integration_weight_labels[0],
                element_output2="out2",
                target=qua_vars[1],
            ),
            adc_stream=stream,
        )
        return tuple(qua_vars)

    def measure_accumulated(
        self,
        pulse_name: str,
        amplitude_scale: Optional[Union[ScalarFloat, Sequence[ScalarFloat]]] = None,
        num_segments: Optional[int] = None,
        segment_length: Optional[int] = None,
        qua_vars: Optional[Tuple[QuaVariableFloat, ...]] = None,
        stream=None,
    ) -> Tuple[QuaVariableFloat, QuaVariableFloat, QuaVariableFloat, QuaVariableFloat]:
        """Perform an accumulated dual demodulation measurement on this channel.

        Instead of two QUA variables (I and Q), this method returns four variables
        (II, IQ, QI, QQ)

        Args:
            pulse_name (str): The name of the pulse to play. Should be registered in
                `self.operations`.
            amplitude_scale (Optional[Union[ScalarFloat, Sequence[ScalarFloat]]]):
                Amplitude scale of the pulse. Can be either a (qua) float, or a list of
                (qua) floats. If None, the pulse is played without amplitude scaling.
            num_segments (int): The number of segments to accumulate.
                Should either specify this or `segment_length`.
            segment_length (int): The length of the segment to accumulate the
                measurement.
                Should either specify this or `num_segments`.
            qua_vars (Tuple[QuaVariableFloat, ...], optional): Four QUA
                variables to store the II, IQ, QI, QQ measurement results.
                If not provided, new variables will be declared and returned.
            stream (Optional[StreamType]): The stream to save the measurement result to.
                If not provided, the raw ADC signal will not be streamed.

        Returns:
            II, IQ, QI, QQ: The QUA variables used to store the measurement results.
                If provided as input, the same variables will be returned.
                If not provided, new variables will be declared and returned.
        """
        pulse: BaseReadoutPulse = self.operations[pulse_name]

        if num_segments is None and segment_length is None:
            raise ValueError(
                "InOutSingleChannel.measure_accumulated requires either 'segment_length' "
                "or 'num_segments' to be provided."
            )
        elif num_segments is not None and segment_length is not None:
            raise ValueError(
                "InOutSingleChannel.measure_accumulated received both 'segment_length' "
                "and 'num_segments'. Please provide only one."
            )
        elif num_segments is None:
            num_segments = int(pulse.length / (4 * segment_length))  # Number of slices
        elif segment_length is None:
            segment_length = int(pulse.length / (4 * num_segments))

        if qua_vars is not None:
            if not isinstance(qua_vars, Sequence) or len(qua_vars) != 4:
                raise ValueError(
                    f"InOutSingleChannel.measure_accumulated received kwarg 'qua_vars' "
                    f"which is not a tuple of four QUA variables. Received {qua_vars=}"
                )
        else:
            qua_vars = [declare(fixed, size=num_segments) for _ in range(4)]

        pulse_name_with_amp_scale = add_amplitude_scale_to_pulse_name(
            pulse_name, amplitude_scale
        )

        integration_weight_labels = list(pulse.integration_weights_mapping)
        measure(
            pulse_name_with_amp_scale,
            self.name,
            demod.accumulated(
                integration_weight_labels[0], qua_vars[0], segment_length, "out1"
            ),
            demod.accumulated(
                integration_weight_labels[1], qua_vars[1], segment_length, "out2"
            ),
            demod.accumulated(
                integration_weight_labels[2], qua_vars[2], segment_length, "out1"
            ),
            demod.accumulated(
                integration_weight_labels[0], qua_vars[3], segment_length, "out2"
            ),
            adc_stream=stream,
        )
        return tuple(qua_vars)

    def measure_sliced(
        self,
        pulse_name: str,
        amplitude_scale: Optional[Union[ScalarFloat, Sequence[ScalarFloat]]] = None,
        num_segments: Optional[int] = None,
        segment_length: Optional[int] = None,
        qua_vars: Optional[Tuple[QuaVariableFloat, ...]] = None,
        stream=None,
    ) -> Tuple[QuaVariableFloat, QuaVariableFloat, QuaVariableFloat, QuaVariableFloat]:
        """Perform a sliced dual demodulation measurement on this channel.

        Instead of two QUA variables (I and Q), this method returns four variables
        (II, IQ, QI, QQ)

        Args:
            pulse_name (str): The name of the pulse to play. Should be registered in
                `self.operations`.
            amplitude_scale (Optional[Union[ScalarFloat, Sequence[ScalarFloat]]]):
                Amplitude scale of the pulse. Can be either a (qua) float, or a list of
                (qua) floats. If None, the pulse is played without amplitude scaling.
            num_segments (int): The number of segments to accumulate.
                Should either specify this or `segment_length`.
            segment_length (int): The length of the segment to accumulate the
                measurement.
                Should either specify this or `num_segments`.
            qua_vars (Tuple[QuaVariableFloat, ...], optional): Four QUA
                variables to store the II, IQ, QI, QQ measurement results.
                If not provided, new variables will be declared and returned.
            stream (Optional[StreamType]): The stream to save the measurement result to.
                If not provided, the raw ADC signal will not be streamed.

        Returns:
            II, IQ, QI, QQ: The QUA variables used to store the measurement results.
                If provided as input, the same variables will be returned.
                If not provided, new variables will be declared and returned.
        """
        pulse: BaseReadoutPulse = self.operations[pulse_name]

        if num_segments is None and segment_length is None:
            raise ValueError(
                "InOutSingleChannel.measure_sliced requires either 'segment_length' "
                "or 'num_segments' to be provided."
            )
        elif num_segments is not None and segment_length is not None:
            raise ValueError(
                "InOutSingleChannel.measure_sliced received both 'segment_length' "
                "and 'num_segments'. Please provide only one."
            )
        elif num_segments is None:
            num_segments = int(pulse.length / (4 * segment_length))  # Number of slices
        elif segment_length is None:
            segment_length = int(pulse.length / (4 * num_segments))

        if qua_vars is not None:
            if not isinstance(qua_vars, Sequence) or len(qua_vars) != 4:
                raise ValueError(
                    f"InOutSingleChannel.measure_sliced received kwarg 'qua_vars' "
                    f"which is not a tuple of four QUA variables. Received {qua_vars=}"
                )
        else:
            qua_vars = [declare(fixed, size=num_segments) for _ in range(4)]

        pulse_name_with_amp_scale = add_amplitude_scale_to_pulse_name(
            pulse_name, amplitude_scale
        )

        integration_weight_labels = list(pulse.integration_weights_mapping)
        measure(
            pulse_name_with_amp_scale,
            self.name,
            demod.sliced(
                integration_weight_labels[0], qua_vars[0], segment_length, "out1"
            ),
            demod.sliced(
                integration_weight_labels[1], qua_vars[1], segment_length, "out2"
            ),
            demod.sliced(
                integration_weight_labels[2], qua_vars[2], segment_length, "out1"
            ),
            demod.sliced(
                integration_weight_labels[0], qua_vars[3], segment_length, "out2"
            ),
            adc_stream=stream,
        )
        return tuple(qua_vars)


@quam_dataclass
class InIQChannel(_InComplexChannel):
    """QUAM component for an IQ input channel

    Args:
        operations (Dict[str, Pulse]): A dictionary of pulses to be played on this
            channel. The key is the pulse label (e.g. "readout") and value is a
            ReadoutPulse.
        id (str, int): The id of the channel, used to generate the name.
            Can be a string, or an integer in which case it will add
            `Channel._default_label`.
        opx_input_I (LF_input_port_types): Channel I input port from the OPX
            perspective, e.g. LFFEMAnalogInputPort("con1", 1, 1)
        opx_input_Q (LF_input_port_types): Channel Q input port from the OPX
            perspective, e.g. LFFEMAnalogInputPort("con1", 1, 2)
        opx_input_offset_I (float, deprecated): The offset of the I channel.
            **Deprecated**: This property has been moved to Port objects. Use
            `OPXPlusAnalogInputPort(offset=...)` on the I port instead.
            See [Port documentation](channel-ports.md) for details.
        opx_input_offset_Q (float, deprecated): The offset of the Q channel.
            **Deprecated**: This property has been moved to Port objects. Use
            `OPXPlusAnalogInputPort(offset=...)` on the Q port instead.
            See [Port documentation](channel-ports.md) for details.
        frequency_converter_down (Optional[FrequencyConverter]): Frequency converter
            QUAM component for the IQ input port. Only needed for the old Octave.
        time_of_flight (int): Round-trip signal duration in nanoseconds.
        smearing (int): Additional window of ADC integration in nanoseconds.
            Used to account for signal smearing.
        input_gain (float): The gain of the input channel. Default is None.
    """

    opx_input_I: LF_input_port_types
    opx_input_Q: LF_input_port_types

    time_of_flight: int = 140
    smearing: int = 0

    opx_input_offset_I: float = None
    opx_input_offset_Q: float = None

    input_gain: Optional[int] = None

    frequency_converter_down: BaseFrequencyConverter = None

    _default_label: ClassVar[str] = "IQ"

    def apply_to_config(self, config: dict):
        """Adds this InOutIQChannel to the QUA configuration.

        See [`QuamComponent.apply_to_config`][quam.core.quam_classes.QuamComponent.apply_to_config]
        for details.
        """
        super().apply_to_config(config)

        # Note outputs instead of inputs because it's w.r.t. the QPU
        element_config = config["elements"][self.name]
        element_config["smearing"] = self.smearing
        element_config["time_of_flight"] = self.time_of_flight

        from quam.components.octave import OctaveDownConverter

        if isinstance(self.frequency_converter_down, OctaveDownConverter):
            octave = self.frequency_converter_down.octave
            if octave is None:
                raise ValueError(
                    f"Error generating config: channel {self.name} has an "
                    f"OctaveDownConverter (id={self.frequency_converter_down.id}) "
                    "without an attached Octave"
                )
            element_config["RF_outputs"] = {
                "port": (octave.name, self.frequency_converter_down.id)
            }
        elif str_ref.is_reference(self.frequency_converter_down):
            raise ValueError(
                f"Error generating config: channel {self.name} could not determine "
                f'"frequency_converter_down", it seems to point to a non-existent '
                f"reference: {self.frequency_converter_down}"
            )
        else:
            # To be filled in next section
            element_config["outputs"] = {}

        # Check for deprecated port properties on channel
        if self.opx_input_offset_I is not None:
            msg = _create_port_property_deprecation_message(
                "InIQChannel",
                "opx_input_offset_I",
                self.opx_input_offset_I,
                self.opx_input_I,
            )
            warnings.warn(msg, DeprecationWarning, stacklevel=2)

        if self.opx_input_offset_Q is not None:
            msg = _create_port_property_deprecation_message(
                "InIQChannel",
                "opx_input_offset_Q",
                self.opx_input_offset_Q,
                self.opx_input_Q,
            )
            warnings.warn(msg, DeprecationWarning, stacklevel=2)

        opx_inputs = [self.opx_input_I, self.opx_input_Q]
        offsets = [self.opx_input_offset_I, self.opx_input_offset_Q]
        input_gain = int(self.input_gain if self.input_gain is not None else 0)
        for k, (opx_input, offset) in enumerate(zip(opx_inputs, offsets), start=1):
            if isinstance(opx_input, LFAnalogInputPort):
                opx_port = opx_input
            elif len(opx_input) == 2:
                opx_port = OPXPlusAnalogInputPort(
                    *opx_input, offset=offset, gain_db=input_gain
                )
                opx_port.apply_to_config(config)
            else:
                opx_port = LFFEMAnalogInputPort(
                    *opx_input, offset=offset, gain_db=input_gain
                )
                opx_port.apply_to_config(config)
            if not isinstance(self.frequency_converter_down, OctaveDownConverter):
                element_config["outputs"][f"out{k}"] = opx_port.port_tuple


@quam_dataclass
class InOutSingleChannel(SingleChannel, InSingleChannel):
    """QUAM component for a single (not IQ) input + output channel.

    Args:
        operations (Dict[str, Pulse]): A dictionary of pulses to be played on this
            channel. The key is the pulse label (e.g. "X90") and value is a Pulse.
        id (str, int): The id of the channel, used to generate the name.
            Can be a string, or an integer in which case it will add
            `Channel._default_label`.
        opx_output (LF_output_port_types): Channel output port from the OPX
            perspective, e.g. LFFEMAnalogOutputPort("con1", 1, 2)
        opx_output_offset (float): DC offset for the output port.
        opx_input (LF_input_port_types): Channel input port from OPX perspective,
            e.g. LFFEMAnalogInputPort("con1", 1, 2)
        opx_input_offset (float): DC offset for the input port.
        filter_fir_taps (List[float]): FIR filter taps for the output port.
        filter_iir_taps (List[float]): IIR filter taps for the output port.
        intermediate_frequency (float): Intermediate frequency of OPX output, default
            is None.
        time_of_flight (int): Round-trip signal duration in nanoseconds.
        smearing (int): Additional window of ADC integration in nanoseconds.
            Used to account for signal smearing.
    """

    pass


@quam_dataclass
class InOutIQChannel(IQChannel, InIQChannel):
    """QUAM component for an IQ channel with both input and output.

    An example of such a channel is a readout resonator, where you may want to
    apply a readout tone and then measure the response.

    Args:
        operations (Dict[str, Pulse]): A dictionary of pulses to be played on this
            channel. The key is the pulse label (e.g. "readout") and value is a
            ReadoutPulse.
        id (str, int): The id of the channel, used to generate the name.
            Can be a string, or an integer in which case it will add
            `Channel._default_label`.
        opx_output_I (LF_output_port_types): Channel I output port from the OPX
            perspective, e.g. LFFEMAnalogOutputPort("con1", 1, 1)
        opx_output_Q (LF_output_port_types): Channel Q output port from the OPX
            perspective, e.g. LFFEMAnalogOutputPort("con1", 1, 2)
        opx_output_offset_I (float): The offset of the I channel. Default is 0.
        opx_output_offset_Q (float): The offset of the Q channel. Default is 0.
        opx_input_I (LF_input_port_types): Channel I input port from the OPX
            perspective, e.g. LFFEMAnalogInputPort("con1", 1, 1)
        opx_input_Q (LF_input_port_types): Channel Q input port from the OPX
            perspective, e.g. LFFEMAnalogInputPort("con1", 1, 2)
        opx_input_offset_I (float): The offset of the I channel. Default is 0.
        opx_input_offset_Q (float): The offset of the Q channel. Default is 0.
        intermediate_frequency (float): Intermediate frequency of the mixer.
            Default is 0.0
        LO_frequency (float): Local oscillator frequency. Default is the LO frequency
            of the frequency converter up component.
        RF_frequency (float): RF frequency of the mixer. By default, the RF frequency
            is inferred by adding the LO frequency and the intermediate frequency.
        frequency_converter_up (FrequencyConverter): Frequency converter QUAM component
            for the IQ output.
        frequency_converter_down (Optional[FrequencyConverter]): Frequency converter
            QUAM component for the IQ input port. Only needed for the old Octave.
        time_of_flight (int): Round-trip signal duration in nanoseconds.
        smearing (int): Additional window of ADC integration in nanoseconds.
            Used to account for signal smearing.
    """

    pass


@quam_dataclass
class InSingleOutIQChannel(IQChannel, InSingleChannel):
    """QUAM component for an IQ output channel with a single input.

    Args:
        operations (Dict[str, Pulse]): A dictionary of pulses to be played on this
            channel. The key is the pulse label (e.g. "readout") and value is a
            ReadoutPulse.
        id (str, int): The id of the channel, used to generate the name.
            Can be a string, or an integer in which case it will add
            `Channel._default_label`.
        opx_output_I (LF_output_port_types): Channel I output port from the OPX
            perspective, e.g. LFFEMAnalogOutputPort("con1", 1, 1)
        opx_output_Q (LF_output_port_types): Channel Q output port from the OPX
            perspective, e.g. LFFEMAnalogOutputPort("con1", 1, 2)
        opx_output_offset_I (float): The offset of the I channel. Default is 0.
        opx_output_offset_Q (float): The offset of the Q channel. Default is 0.
        opx_input (LF_input_port_types): Channel input port from OPX perspective,
            e.g. LFFEMAnalogInputPort("con1", 1, 1)
        opx_input_offset (float): DC offset for the input port.
        intermediate_frequency (float): Intermediate frequency of the mixer.
            Default is 0.0
        LO_frequency (float): Local oscillator frequency. Default is the LO frequency
            of the frequency converter up component.
        RF_frequency (float): RF frequency of the mixer. By default, the RF frequency
            is inferred by adding the LO frequency and the intermediate frequency.
        frequency_converter_up (FrequencyConverter): Frequency converter QUAM component
            for the IQ output.
        time_of_flight (int): Round-trip signal duration in nanoseconds.
        smearing (int): Additional window of ADC integration in nanoseconds.
            Used to account for signal smearing.
    """

    pass


@quam_dataclass
class InIQOutSingleChannel(SingleChannel, InIQChannel):
    """QUAM component for an IQ input channel with a single output.

    Args:
        operations (Dict[str, Pulse]): A dictionary of pulses to be played on this
            channel. The key is the pulse label (e.g. "readout") and value is a
            ReadoutPulse.
        id (str, int): The id of the channel, used to generate the name.
            Can be a string, or an integer in which case it will add
            `Channel._default_label`.
        opx_output (LF_output_port_types): Channel output port from the OPX
            perspective, e.g. LFFEMAnalogOutputPort("con1", 1, 1)
        opx_output_offset (float): DC offset for the output port.
        opx_input_I (LF_input_port_types): Channel I input port from the OPX
            perspective, e.g. LFFEMAnalogInputPort("con1", 1, 1)
        opx_input_Q (LF_input_port_types): Channel Q input port from the OPX
            perspective, e.g. LFFEMAnalogInputPort("con1", 1, 2)
        opx_input_offset_I (float): The offset of the I channel. Default is 0.
        opx_input_offset_Q (float): The offset of the Q channel. Default is 0.
        filter_fir_taps (List[float]): FIR filter taps for the output port.
        filter_iir_taps (List[float]): IIR filter taps for the output port.
        intermediate_frequency (float): Intermediate frequency of OPX output, default
            is None.
        time_of_flight (int): Round-trip signal duration in nanoseconds.
        smearing (int): Additional window of ADC integration in nanoseconds.
            Used to account for signal smearing.
    """

    pass


@quam_dataclass
class MWChannel(_OutComplexChannel):
    """QUAM component for a MW FEM output channel

    Args:
        operations (Dict[str, Pulse]): A dictionary of pulses to be played on this
            channel. The key is the pulse label (e.g. "X90") and value is a Pulse.
        id (str, int): The id of the channel, used to generate the name.
            Can be a string, or an integer in which case it will add
            `Channel._default_label`.
        opx_output (MWFEMAnalogOutputPort): Channel output port from the OPX perspective.
        intermediate_frequency (float): Intermediate frequency of OPX output, default
            is None.
        upconverter (int): The upconverter to use. Default is 1.
        time_of_flight (int): Round-trip signal duration in nanoseconds.
        smearing (int): Additional window of ADC integration in nanoseconds.
            Used to account for signal smearing.
    """

    opx_output: MWFEMAnalogOutputPort
    upconverter: int = 1

    LO_frequency: float = "#./upconverter_frequency"
    RF_frequency: float = "#./inferred_RF_frequency"

    def apply_to_config(self, config: Dict) -> None:
        super().apply_to_config(config)

        element_config = config["elements"][self.name]
        element_config["MWInput"] = {
            "port": self.opx_output.port_tuple,
            "upconverter": self.upconverter,
        }

    @property
    def upconverter_frequency(self) -> float:
        """Determine the upconverter frequency from the opx_output.

        If the upconverter frequency is not set, the upconverter frequency is inferred
        from the upconverters dictionary.

        Returns:
            The upconverter frequency.

        Raises:
            ValueError: If the upconverter frequency is not set and cannot be inferred.
        """
        if self.opx_output.upconverter_frequency is not None:
            return self.opx_output.upconverter_frequency
        if self.opx_output.upconverters is not None:
            upconverter_config = self.opx_output.upconverters.get(self.upconverter)
            if upconverter_config is None:
                raise ValueError(
                    f"MWChannel: Upconverter {self.upconverter} not found in upconverters dictionary"
                )
            if "frequency" not in upconverter_config:
                raise ValueError(
                    f"MWChannel: 'frequency' key not found in upconverter {self.upconverter} configuration"
                )
            return upconverter_config["frequency"]
        raise ValueError(
            "MWChannel: Either upconverter_frequency or upconverters must be provided"
        )


@quam_dataclass
class InMWChannel(_InComplexChannel):
    """QUAM component for a MW FEM input channel

    Args:
        operations (Dict[str, Pulse]): A dictionary of pulses to be played on this
            channel. The key is the pulse label (e.g. "X90") and value is a Pulse.
        id (str, int): The id of the channel, used to generate the name.
            Can be a string, or an integer in which case it will add
            `Channel._default_label`.
        opx_input (MWFEMAnalogInputPort): Channel input port from the OPX
            perspective, e.g. MWFEMAnalogInputPort("con1", 1, 1)
        intermediate_frequency (float): Intermediate frequency of OPX output, default
            is None.
        time_of_flight (int): Round-trip signal duration in nanoseconds. Default is 280,
            which is a reasonable default for the MW FEM.
        smearing (int): Additional window of ADC integration in nanoseconds.
            Used to account for signal smearing. Default is 0.
    """

    opx_input: MWFEMAnalogInputPort

    time_of_flight: int = 280
    smearing: int = 0

    def apply_to_config(self, config: Dict) -> None:
        super().apply_to_config(config)

        element_config = config["elements"][self.name]
        element_config["MWOutput"] = {"port": self.opx_input.port_tuple}
        element_config["smearing"] = self.smearing
        element_config["time_of_flight"] = self.time_of_flight


@quam_dataclass
class InOutMWChannel(MWChannel, InMWChannel):
    """QUAM component for a MW FEM input channel

    Args:
        operations (Dict[str, Pulse]): A dictionary of pulses to be played on this
            channel. The key is the pulse label (e.g. "X90") and value is a Pulse.
        id (str, int): The id of the channel, used to generate the name.
            Can be a string, or an integer in which case it will add
            `Channel._default_label`.
        opx_output (MWFEMAnalogOutputPort): Channel output port from the OPX
            perspective, e.g. MWFEMAnalogOutputPort("con1", 1, 1)
        opx_input (MWFEMAnalogInputPort): Channel input port from the OPX
            perspective, e.g. MWFEMAnalogInputPort("con1", 1, 1)
        intermediate_frequency (float): Intermediate frequency of OPX output, default
            is None.
        upconverter (int): The upconverter to use. Default is 1.
        time_of_flight (int): Round-trip signal duration in nanoseconds.
        smearing (int): Additional window of ADC integration in nanoseconds.
            Used to account for signal smearing.
    """

    pass



import pytest

from quam.core import *


@quam_dataclass
class BasicQuamComponent(QuamComponent):
    a: int = 4


@quam_dataclass
class BareQuamRoot(QuamRoot):
    required_component: BasicQuamComponent
    required_component_2: BasicQuamComponent
    optional_component: BasicQuamComponent = None


@pytest.fixture
def machine():
    return BareQuamRoot(
        required_component=BasicQuamComponent(a="#/required_component_2/a"),
        required_component_2=BasicQuamComponent(a=43),
    )


def test_overwrite_reference_error(machine):
    assert machine.required_component.a == 43

    with pytest.raises(ValueError):
        machine.required_component.a = 42


def test_overwrite_reference_after_None(machine):
    assert machine.required_component.a == 43

    machine.required_component.a = None

    assert machine.required_component.a is None

    machine.required_component.a = 42

    assert machine.required_component.a == 42


def test_overwrite_reference_to_reference(machine):
    machine.required_component.a = "#/required_component_2/a"
    assert machine.required_component.a == 43

    machine.required_component.a = "#/required_component_2"
    assert machine.required_component.a == machine.required_component_2


def test_overwrite_value_to_reference(machine):
    assert machine.required_component_2.a == 43
    machine.required_component_2.a = "#/required_component_2"
    assert machine.required_component_2.a == machine.required_component_2


def test_overwrite_nonexistent_to_value(machine):
    machine.required_component.b = 43


def test_overwrite_nonexistent_to_ref(machine):
    machine.required_component.b = "#/required_component_2/a"
    assert machine.required_component.b == 43


def test_overwrite_nonexistent_to_None(machine):
    machine.required_component.b = None
    assert machine.required_component.b is None


def test_dict_overwrite_ref(machine):
    d = QuamDict(a="#./b", b=42)
    assert d._initialized
    assert d.a == 42

    with pytest.raises(ValueError):
        d.a = 43

    with pytest.raises(ValueError):
        d["a"] = 43



import pytest

from quam.components import *
from quam.examples.superconducting_qubits.components import *
from quam.components.channels import IQChannel
from quam.core import QuamRoot, quam_dataclass


def test_basic_transmon():
    transmon = Transmon(id=1)
    assert transmon.name == "q1"
    assert transmon.id == 1
    assert transmon.xy is None
    assert transmon.z is None

    transmon.id = "Q1"
    assert transmon.name == "Q1"


def test_transmon_xy():
    transmon = Transmon(
        id=1,
        xy=IQChannel(
            opx_output_I=("con1", 1),
            opx_output_Q=("con1", 2),
            frequency_converter_up=FrequencyConverter(
                mixer=Mixer(),
                local_oscillator=LocalOscillator(frequency=5e9),
            ),
            intermediate_frequency=100e6,
        ),
    )

    assert transmon.xy.name == "q1.xy"
    assert transmon.xy.mixer.name == "q1.xy.mixer"
    assert not transmon.xy.operations
    assert transmon.xy.mixer.intermediate_frequency == 100e6
    assert transmon.z is None

    cfg = {"controllers": {}, "elements": {}}

    # References first have to be set to None
    with pytest.raises(ValueError):
        transmon.xy.mixer.local_oscillator_frequency = 5e9
    transmon.xy.mixer.local_oscillator_frequency = None
    transmon.xy.mixer.local_oscillator_frequency = 5e9

    assert transmon.xy.rf_frequency == 5.1e9

    transmon.xy.apply_to_config(cfg)
    expected_cfg = {
        "elements": {
            "q1.xy": {
                "mixInputs": {
                    "I": ("con1", 1),
                    "Q": ("con1", 2),
                    "lo_frequency": 5000000000.0,
                    "mixer": "q1.xy.mixer",
                },
                "intermediate_frequency": 100e6,
                "operations": {},
            },
        },
        "controllers": {
            "con1": {
                "analog_outputs": {
                    1: {"delay": 0, "shareable": False},
                    2: {"delay": 0, "shareable": False},
                },
            }
        },
    }
    assert cfg == expected_cfg


def test_transmon_xy_opx1000():
    transmon = Transmon(
        id=1,
        xy=IQChannel(
            opx_output_I=("con1", 2, 1),
            opx_output_Q=("con1", 2, 2),
            frequency_converter_up=FrequencyConverter(
                mixer=Mixer(),
                local_oscillator=LocalOscillator(frequency=5e9),
            ),
            intermediate_frequency=100e6,
        ),
    )

    assert transmon.xy.name == "q1.xy"
    assert transmon.xy.mixer.name == "q1.xy.mixer"
    assert not transmon.xy.operations
    assert transmon.xy.mixer.intermediate_frequency == 100e6
    assert transmon.z is None

    cfg = {"controllers": {}, "elements": {}}

    # References first have to be set to None
    with pytest.raises(ValueError):
        transmon.xy.mixer.local_oscillator_frequency = 5e9
    transmon.xy.mixer.local_oscillator_frequency = None
    transmon.xy.mixer.local_oscillator_frequency = 5e9

    assert transmon.xy.rf_frequency == 5.1e9

    transmon.xy.apply_to_config(cfg)
    expected_cfg = {
        "elements": {
            "q1.xy": {
                "mixInputs": {
                    "I": ("con1", 2, 1),
                    "Q": ("con1", 2, 2),
                    "lo_frequency": 5000000000.0,
                    "mixer": "q1.xy.mixer",
                },
                "intermediate_frequency": 100e6,
                "operations": {},
            },
        },
        "controllers": {
            "con1": {
                "fems": {
                    2: {
                        "type": "LF",
                        "analog_outputs": {
                            1: {
                                "delay": 0,
                                "shareable": False,
                                "sampling_rate": 1e9,
                                "upsampling_mode": "mw",
                                "output_mode": "direct",
                            },
                            2: {
                                "delay": 0,
                                "shareable": False,
                                "sampling_rate": 1e9,
                                "upsampling_mode": "mw",
                                "output_mode": "direct",
                            },
                        },
                    }
                }
            }
        },
    }
    assert cfg == expected_cfg


def test_transmon_add_pulse():
    transmon = Transmon(
        id=1,
        xy=IQChannel(
            opx_output_I=("con1", 1),
            opx_output_Q=("con1", 2),
            frequency_converter_up=FrequencyConverter(
                mixer=Mixer(),
                local_oscillator=LocalOscillator(frequency=5e9),
            ),
            intermediate_frequency=100e6,
        ),
    )
    transmon.xy.operations["X180"] = pulses.DragPulse(
        amplitude=1, sigma=4, alpha=2, anharmonicity=200e6, length=20, axis_angle=0
    )

    quam_dict = transmon.to_dict(include_defaults=False)
    expected_quam_dict = {
        "__class__": "quam.examples.superconducting_qubits.components.Transmon",
        "id": 1,
        "xy": {
            "__class__": "quam.components.channels.IQChannel",
            "intermediate_frequency": 100000000.0,
            "opx_output_I": ("con1", 1),
            "opx_output_Q": ("con1", 2),
            "frequency_converter_up": {
                "__class__": "quam.components.hardware.FrequencyConverter",
                "mixer": {"__class__": "quam.components.hardware.Mixer"},
                "local_oscillator": {
                    "frequency": 5000000000.0,
                    "__class__": "quam.components.hardware.LocalOscillator",
                },
            },
            "operations": {
                "X180": {
                    "__class__": "quam.components.pulses.DragPulse",
                    "amplitude": 1,
                    "sigma": 4,
                    "alpha": 2,
                    "anharmonicity": 200000000.0,
                    "axis_angle": 0.0,
                    "length": 20,
                }
            },
        },
    }
    assert quam_dict == expected_quam_dict

    config = {"controllers": {}, "elements": {}, "pulses": {}, "waveforms": {}}
    transmon.xy.apply_to_config(config)

    assert config == {
        "controllers": {
            "con1": {
                "analog_outputs": {
                    1: {
                        "delay": 0,
                        "shareable": False,
                    },
                    2: {
                        "delay": 0,
                        "shareable": False,
                    },
                },
            }
        },
        "elements": {
            "q1.xy": {
                "mixInputs": {
                    "I": ("con1", 1),
                    "Q": ("con1", 2),
                    "lo_frequency": 5000000000.0,
                    "mixer": "q1.xy.mixer",
                },
                "intermediate_frequency": 100e6,
                "operations": {"X180": "q1.xy.X180.pulse"},
            },
        },
        "pulses": {},
        "waveforms": {},
    }

    config = {"controllers": {}, "elements": {}, "pulses": {}, "waveforms": {}}
    transmon.xy.operations["X180"].apply_to_config(config)

    from qualang_tools.config.waveform_tools import drag_gaussian_pulse_waveforms

    I, Q = drag_gaussian_pulse_waveforms(
        amplitude=1, sigma=4, alpha=2, anharmonicity=200e6, length=20
    )
    expected_cfg = {
        "controllers": {},
        "elements": {},
        "pulses": {
            "q1.xy.X180.pulse": {
                "operation": "control",
                "length": 20,
                "waveforms": {"I": "q1.xy.X180.wf.I", "Q": "q1.xy.X180.wf.Q"},
            },
        },
        "waveforms": {
            "q1.xy.X180.wf.I": {"type": "arbitrary", "samples": I},
            "q1.xy.X180.wf.Q": {"type": "arbitrary", "samples": Q},
        },
    }
    assert config == expected_cfg


@quam_dataclass
class QuamTestSingle(QuamRoot):
    qubit: Transmon


quam_dict_single = {"qubit": {"id": 0}}


quam_dict_single_nested = {
    "qubit": {
        "id": 0,
        "xy": {
            "opx_output_I": ("con1", 0),
            "opx_output_Q": ("con1", 1),
            "intermediate_frequency": 100e6,
            "frequency_converter_up": {
                "mixer": {},
                "local_oscillator": {"power": 10, "frequency": 6e9},
            },
        },
    }
}


def test_instantiation_single_element():
    quam = QuamTestSingle.load(quam_dict_single)

    assert isinstance(quam.qubit, Transmon)
    assert quam.qubit.id == 0
    assert quam.qubit.xy is None

    assert quam.qubit.get_root() is quam


def test_instantiation_single_nested_element():
    quam = QuamTestSingle.load(quam_dict_single_nested)

    assert quam.qubit.xy.mixer.name == "q0.xy.mixer"
    assert quam.qubit.xy.mixer.local_oscillator_frequency == 6e9

    assert quam.qubit.get_root() is quam
    assert quam.qubit.xy.get_root() is quam
    assert quam.qubit.xy.mixer.get_root() is quam



from dataclasses import dataclass
from typing import List, Union
from quam.core import *
from quam.core.quam_instantiation import *
from quam.utils import get_full_class_path


@quam_dataclass
class QuamComponentTest(QuamComponent):
    str_val: str


def test_instantiate_from_class():
    quam_component = QuamComponentTest(str_val="hi")
    loaded_component = instantiate_quam_class(
        QuamComponentTest, quam_component.to_dict()
    )

    assert isinstance(loaded_component, QuamComponentTest)
    assert loaded_component.to_dict() == quam_component.to_dict()

    d = quam_component.to_dict()
    d["__class__"] = get_full_class_path(QuamComponentTest)
    loaded_explicit_component = instantiate_quam_class(QuamComponent, d)
    loaded_explicit_component2 = instantiate_quam_class(QuamComponent, d)
    assert type(loaded_explicit_component) == type(loaded_explicit_component2)

    assert isinstance(loaded_explicit_component, QuamComponentTest)


def test_instantiate_nondefault_list_from_dict():
    @quam_dataclass
    class QuamBasicComponent(QuamComponent):
        l: Union[int, List[int]] = 42

    quam_component = QuamBasicComponent(l=[1, 2, 3])

    d = quam_component.to_dict()
    instantiate_quam_class(QuamBasicComponent, d)


@quam_dataclass
class QuamBasicComponent(QuamComponent):
    int_val: int = 42


@quam_dataclass
class QuamOuterComponent(QuamComponent):
    list_basic_components: Union[int, List[QuamBasicComponent]]


def test_instantiate_explicit_class_with_union_type():
    quam_component = QuamOuterComponent(
        list_basic_components=[QuamBasicComponent(int_val=42)]
    )

    d = quam_component.to_dict()
    component = instantiate_quam_class(QuamOuterComponent, d)

    assert isinstance(component, QuamOuterComponent)
    assert isinstance(component.list_basic_components[0], QuamBasicComponent)
    assert component.to_dict() == quam_component.to_dict()


@quam_dataclass
class Component1(QuamComponent):
    pass

@quam_dataclass
class Component2(QuamComponent):
    pass


def test_instantiation_unknown_error():


    instantiate_attr(
        attr_val={"a": {"__class__": get_full_class_path(Component1)}},
        expected_type=Dict[str, Union[Component1, Component2]],
        fix_attrs=True,
        validate_type=True,
        str_repr="hi"
    )



from quam.core import *
from quam.utils import get_full_class_path


def test_get_transmon_class_path():
    from quam.examples.superconducting_qubits.components import Transmon

    class_path = get_full_class_path(Transmon)
    assert class_path == "quam.examples.superconducting_qubits.components.Transmon"


@quam_dataclass
class QuamComponentTest(QuamComponent):
    str_val: str


def test_get_local_class_path():
    class_path = get_full_class_path(QuamComponentTest)
    assert class_path == "test_get_full_class_path.QuamComponentTest"
